import logging
from typing import cast

from aiohttp import web

from glQiwiApi.core.event_fetching.webhooks.views.base import BaseWebhookView
from glQiwiApi.qiwi.clients.p2p.types import BillWebhook
from glQiwiApi.types.exceptions import WebhookSignatureUnverifiedError

logger = logging.getLogger('glQiwiApi.webhooks.p2p')


class QiwiBillWebhookView(BaseWebhookView[BillWebhook]):
    """View, which processes p2p notifications"""

    def _validate_event_signature(self, update: BillWebhook) -> None:
        sha256_signature = cast(
            str, self.request.headers.get('X-Api-Signature-SHA256')
        )  # pragma: no cover

        try:  # pragma: no cover
            update.verify_signature(sha256_signature, self._encryption_key)  # pragma: no cover
        except WebhookSignatureUnverifiedError:  # pragma: no cover
            logger.debug(
                'Blocking request due to invalid signature of payload.'
            )  # pragma: no cover
            raise web.HTTPBadRequest()  # pragma: no cover

    async def ok_response(self) -> web.Response:
        return web.json_response(data={'error': '0'})



from .account_info import UserProfile
from .balance import Balance
from .commission import Commission
from .identification import Identification
from .limit import Limit
from .other import CrossRate, PaymentDetails, PaymentMethod
from .partner import Partner
from .payment_info import PaymentInfo, QiwiPayment
from .qiwi_master import Card, OrderDetails
from .restriction import Restriction
from .stats import Statistic
from .transaction import History, Source, Transaction, TransactionStatus, TransactionType
from .webhooks import TransactionWebhook, WebhookInfo

__all__ = (
    'UserProfile',
    'Transaction',
    'Statistic',
    'Limit',
    'Balance',
    'Identification',
    'PaymentInfo',
    'OrderDetails',
    'Partner',
    'TransactionWebhook',
    'WebhookInfo',
    'CrossRate',
    'PaymentMethod',
    'Card',
    'Restriction',
    'Commission',
    'TransactionType',
    'QiwiPayment',
    'TransactionStatus',
    'Source',
    'PaymentDetails',
    'History',
)



from __future__ import annotations

import abc
import io
import os
import pathlib
from types import TracebackType
from typing import Any, BinaryIO, Generic, Optional, Type, TypeVar

InputType = TypeVar('InputType')

__all__ = ('AbstractInput', 'PlainPathInput', 'PathlibPathInput', 'BinaryIOInput')


class AbstractInput(abc.ABC, Generic[InputType]):
    def __init__(self, input_: InputType) -> None:
        self._input = input_
        self._file_descriptor: Optional[BinaryIO] = None

    @abc.abstractmethod
    def get_file(self) -> BinaryIO:
        ...

    def get_path(self) -> str:
        raise TypeError(
            f"{self.__class__.__qualname__} doesn't provide a mechanism to get path to file"
        )

    def get_filename(self) -> str:
        raise TypeError(
            f"{self.__class__.__qualname__} doesn't provide a mechanism to get filename"
        )

    def close(self) -> None:
        if self._file_descriptor is None:
            return None
        self._file_descriptor.close()

    def __enter__(self) -> AbstractInput[Any]:
        self._file_descriptor = self.get_file()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()


class PlainPathInput(AbstractInput[str]):
    def get_file(self) -> BinaryIO:
        if pathlib.Path(self._input).is_file() is False:
            raise TypeError(f'Input {self._input} is not a file!')
        descriptor = open(self._input, 'rb')
        self._file_descriptor = descriptor
        return descriptor

    def get_path(self) -> str:
        return self._input

    def get_filename(self) -> str:
        return os.path.split(self._input)[-1]


class PathlibPathInput(AbstractInput[pathlib.Path]):
    def get_file(self) -> BinaryIO:
        descriptor = open(self._input, 'rb')
        self._file_descriptor = descriptor
        return descriptor

    def get_path(self) -> str:
        return str(self._input.resolve())

    def get_filename(self) -> str:
        return self._input.name


class BinaryIOInput(AbstractInput[BinaryIO]):
    def get_file(self) -> BinaryIO:
        return self._input

    @classmethod
    def from_bytes(cls: Type[BinaryIOInput], b: bytes) -> BinaryIOInput:
        return cls(input_=io.BytesIO(b))



from __future__ import annotations

import enum
from datetime import datetime
from typing import Any, Dict, Iterator, List, Optional, Union, cast

from pydantic import Field

from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.base import Base


class TransactionType(str, enum.Enum):
    IN = 'IN'
    OUT = 'OUT'
    ALL = 'ALL'
    QIWI_CARD = 'QIWI_CARD'


class TransactionStatus(str, enum.Enum):
    WAITING = 'WAITING'
    SUCCESS = 'SUCCESS'
    ERROR = 'ERROR'


class Source(str, enum.Enum):
    RUB = 'QW_RUB'
    USD = 'QW_USD'
    EUR = 'QW_EUR'
    CARD = 'CARD'
    MK = 'MK'


class Provider(Base):
    """object: Provider"""

    id: Optional[int] = None
    """ID провайдера в QIWI Wallet"""

    short_name: Optional[str] = Field(default=None, alias='shortName')
    """краткое наименование провайдера"""

    long_name: Optional[str] = Field(alias='longName', default=None)
    """развернутое наименование провайдера"""

    logo_url: Optional[str] = Field(alias='logoUrl', default=None)
    """ссылка на логотип провайдера"""

    description: Optional[str] = None
    """описание провайдера (HTML)"""

    keys: Optional[Union[str, List[Any]]] = None
    """список ключевых слов"""

    site_url: Optional[str] = Field(alias='siteUrl', default=None)
    """сайт провайдера"""


class Transaction(Base):
    """object: Transaction"""

    id: int = Field(alias='txnId')
    """	ID транзакции в сервисе QIWI Кошелек"""

    person_id: int = Field(alias='personId')
    """Номер кошелька"""

    date: datetime
    """
    Для запросов истории платежей - Дата/время платежа,
    во временной зоне запроса (см. параметр startDate).
    Для запросов данных о транзакции - Дата/время платежа, время московское
    """

    status: TransactionStatus
    """
    Статус платежа. Возможные значения:
    WAITING - платеж проводится,
    SUCCESS - успешный платеж,
    ERROR - ошибка платежа.
    """

    type: TransactionType
    """
    Тип платежа. Возможные значения:
    IN - пополнение,
    OUT - платеж,
    QIWI_CARD - платеж с карт QIWI (QVC, QVP).
    """

    status_text: str = Field(alias='statusText')
    """Текстовое описание статуса платежа"""

    trm_transaction_id: str = Field(alias='trmTxnId')
    """Клиентский ID транзакции"""

    to_account: str = Field(alias='account')
    """
    Для платежей - номер счета получателя.
    Для пополнений - номер отправителя,
    терминала или название агента пополнения кошелька
    """
    sum: AmountWithCurrency
    """Данные о сумме платежа или пополнения."""

    commission: AmountWithCurrency
    """Данные о комиссии"""

    total: AmountWithCurrency
    """Общие данные о платеже в формате объекта Sum"""

    provider: Provider
    """Данные о провайдере."""

    source: Optional[Provider] = None
    """Служебная информация"""

    comment: Optional[str] = None
    """	Комментарий к платежу"""

    currency_rate: int = Field(alias='currencyRate')
    """Курс конвертации (если применяется в транзакции)"""

    _extras: Optional[Dict[Any, Any]] = Field(None, alias='extras')
    """Служебная информация"""

    _cheque_ready: bool = Field(False, alias='chequeReady')
    """Специальное поле"""

    _bank_document_available: bool = Field(False, alias='bankDocumentAvailable')
    """Специальное поле"""

    _repeat_payment_enabled: bool = Field(False, alias='repeatPaymentEnabled')
    """Специальное поле"""

    _favorite_payment_enabled: bool = Field(False, alias='favoritePaymentEnabled')
    """Специальное поле"""

    _regular_payment_enabled: bool = Field(False, alias='regularPaymentEnabled')
    """Специальное поле"""

    class Config:
        use_enum_values = True


class History(Base):
    transactions: List[Transaction] = Field(..., alias='data')
    next_transaction_date: Optional[datetime] = Field(None, alias='nextTxnDate')
    next_transaction_id: Optional[int] = Field(None, alias='nextTxnId')

    def __iter__(self) -> Iterator[Transaction]:  # type: ignore
        for t in self.transactions:
            yield t

    def __len__(self) -> int:
        return len(self.transactions)

    def __str__(self) -> str:
        if not self.transactions:
            return 'Empty history'
        start_txn_date = self.transactions[0].date
        end_txn_date = self.transactions[-1].date
        return (
            f'History from {start_txn_date} to {end_txn_date} | '
            f'next transaction id = {self.next_transaction_id} | size = {len(self)}'
        )

    def sorted_by_id(self) -> History:
        return self.copy(
            exclude={'transactions'},
            update=dict(
                transactions=sorted(self.transactions, key=lambda txn: txn.id)  # type: ignore
            ),
        )

    def sorted_by_date(self, *, from_latest_to_earliest: bool = False) -> History:
        return self.copy(
            exclude={'transactions'},
            update=dict(
                transactions=sorted(
                    self.transactions,
                    key=lambda txn: txn.date,  # type: ignore
                    reverse=from_latest_to_earliest,
                )
            ),
        )

    def __getitem__(self, i: Any) -> Transaction:
        return cast(Transaction, self.transactions.__getitem__(i))

    def __bool__(self) -> bool:
        return bool(self.transactions)

    def first(self) -> Transaction:
        return self.transactions[0]

    def last(self) -> Transaction:
        return self.transactions[-1]



from __future__ import annotations

import abc
from typing import Optional

from glQiwiApi.qiwi.clients.wallet.types.webhooks import TransactionWebhook, WebhookPayment

from .base import Handler


class AbstractTransactionWebhookHandler(Handler[TransactionWebhook], abc.ABC):
    @property
    def hook_id(self) -> str:
        return self.event.id

    @property
    def payment(self) -> Optional[WebhookPayment]:
        return self.event.payment



from __future__ import annotations

import typing as t

from glQiwiApi.plugins import Pluggable

if t.TYPE_CHECKING:
    from glQiwiApi.utils.compat import Dispatcher


class AiogramPollingPlugin(Pluggable):
    """
    Builtin telegram proxy.
    Allows you to use Telegram and QIWI webhooks together

    """

    def __init__(
        self,
        dispatcher: Dispatcher,
        timeout: int = 20,
        relax: float = 0.1,
        limit: t.Optional[t.Any] = None,
        reset_webhook: t.Optional[t.Any] = None,
        fast: t.Optional[bool] = True,
        error_sleep: int = 5,
        allowed_updates: t.Optional[t.List[str]] = None,
    ) -> None:
        self._allowed_updates = allowed_updates
        self._error_sleep = error_sleep
        self._fast = fast
        self._reset_webhook = reset_webhook
        self._timeout = timeout
        self._relax = relax
        self._limit = limit
        self._dispatcher = dispatcher

    async def install(self, ctx: t.Dict[t.Any, t.Any]) -> None:
        """
        Set up polling to run polling qiwi updates concurrently with aiogram

        :param ctx: you can pass on loop as key/value parameter
        """
        await self._dispatcher.start_polling(
            timeout=self._timeout,
            reset_webhook=self._reset_webhook,
            relax=self._relax,
            allowed_updates=self._allowed_updates,
            limit=self._limit,
            error_sleep=self._error_sleep,
            fast=self._fast,
        )

    async def shutdown(self) -> None:
        self._dispatcher.stop_polling()



from typing import ClassVar, List

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Partner


class GetPartners(QiwiAPIMethod[List[Partner]]):
    url: ClassVar[str] = 'http://edge.qiwi.com/locator/v3/ttp-groups'
    http_method: ClassVar[str] = 'GET'



from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod


class UnblockQiwiMasterCard(QiwiAPIMethod[Dict[str, Any]]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/cards/v2/persons/{phone_number}/cards/{card_id}/unblock'
    http_method: ClassVar[str] = 'GET'

    card_id: str = Field(..., path_runtime_value=True)
    phone_number: str = Field(..., path_runtime_value=True)



import typing

from glQiwiApi.core.abc.base_api_client import BaseAPIClient, RequestServiceFactoryType
from glQiwiApi.core.request_service import RequestService, RequestServiceProto
from glQiwiApi.core.session import AiohttpSessionHolder
from glQiwiApi.qiwi.clients.maps.methods.get_partners import GetPartners
from glQiwiApi.qiwi.clients.maps.methods.get_terminals import GetTerminals
from glQiwiApi.qiwi.clients.maps.types.polygon import Polygon
from glQiwiApi.qiwi.clients.maps.types.terminal import Terminal
from glQiwiApi.qiwi.clients.wallet.types import Partner


class QiwiMaps(BaseAPIClient):
    """
    QIWI Terminal Maps API allows you to locate
    QIWI terminals on the territory of the Russian Federation

    """

    def __init__(
        self,
        request_service_factory: typing.Optional[RequestServiceFactoryType] = None,
    ) -> None:
        super().__init__(request_service_factory)

    async def _create_request_service(self) -> RequestServiceProto:
        return RequestService(
            session_holder=AiohttpSessionHolder(
                headers={
                    'Content-type': 'application/json',
                    'Accept': 'application/json',
                }
            )
        )

    async def terminals(
        self,
        polygon: Polygon,
        zoom: typing.Optional[int] = None,
        pop_if_inactive_x_mins: int = 30,
        include_partners: typing.Optional[bool] = None,
        partners_ids: typing.Optional[typing.List[typing.Any]] = None,
        cache_terminals: typing.Optional[bool] = None,
        card_terminals: typing.Optional[bool] = None,
        identification_types: typing.Optional[int] = None,
        terminal_groups: typing.Optional[typing.List[typing.Any]] = None,
    ) -> typing.List[Terminal]:
        """
        Get map of terminals sent for passed polygon with additional params

        :param polygon: glQiwiApi.types.Polygon object
        :param zoom:
         https://tech.yandex.ru/maps/doc/staticapi/1.x/dg/concepts/map_scale-docpage/
        :param pop_if_inactive_x_mins: do not show if terminal
         was inactive for X minutes default 0.5 hours
        :param include_partners: result will include/exclude partner terminals
        :param partners_ids: Not fixed IDS array look at docs
        :param cache_terminals: result will include or exclude
         cache-acceptable terminals
        :param card_terminals: result will include or
         exclude card-acceptable terminals
        :param identification_types: `0` - not identified, `1` -
         partly identified, `2` - fully identified
        :param terminal_groups: look at QiwiMaps.partners
        :return: list of Terminal instances
        """
        return await self._request_service.execute_api_method(
            GetTerminals(
                polygon=polygon,
                zoom=zoom,
                pop_if_inactive_x_mins=pop_if_inactive_x_mins,
                include_partners=include_partners,
                partners_ids=partners_ids,
                cache_terminals=cache_terminals,
                card_terminals=card_terminals,
                identification_types=identification_types,
                terminal_groups=terminal_groups,
            )
        )

    async def partners(self) -> typing.List[Partner]:
        """
        Get terminal partners for ttpGroups
        :return: list of TTPGroups
        """
        return await self._request_service.execute_api_method(GetPartners())



from typing import Optional, Union

from glQiwiApi.types import _currencies
from glQiwiApi.types.amount import CurrencyModel


class Currency:
    """
        Class with many currencies
    import glQiwiApi.types.basics    >>> usd = Currency.get('840')
        >>> usd
        ... CurrencyModel(code='USD', decimal_digits=2, name='US Dollar', name_plural='US dollars',
        ...                   rounding=0, symbol='$', symbol_native='$')

        >>> usd.symbol
        ... '$'
    """

    @classmethod
    def get(cls, currency_code: Union[str, int]) -> Optional[CurrencyModel]:
        """
        Implements class-based getitem behaviour

        >>> Currency.get('840').symbol
        ... '$'
        >>> Currency.get('USD').symbol
        ... '$'
        :param currency_code: ISO 4217 string or CODE
        :return: Currency object
        """
        if isinstance(currency_code, int) or currency_code.isdigit():
            return _currencies.described.get(_currencies.codes_number[str(currency_code)])
        else:
            return _currencies.described.get(currency_code.upper())



from __future__ import annotations

import base64
import hashlib
import hmac
from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseConfig, Extra, Field

from glQiwiApi.types.amount import HashablePlainAmount, PlainAmount
from glQiwiApi.types.base import HashableBase
from glQiwiApi.types.exceptions import WebhookSignatureUnverifiedError


class Customer(HashableBase):
    phone_number: Optional[str] = Field(None, alias='phone')
    email: Optional[str] = None
    account: Optional[str] = None


class BillStatus(HashableBase):
    value: str
    changed_datetime: Optional[datetime] = Field(None, alias='changedDateTime')


class CustomFields(HashableBase):
    pay_sources_filter: Optional[str] = Field(None, alias='paySourcesFilter')
    theme_code: Optional[str] = Field(None, alias='themeCode')


class BillError(HashableBase):
    service_name: str = Field(..., alias='serviceName')
    error_code: str = Field(..., alias='errorCode')
    description: str
    user_message: str = Field(..., alias='userMessage')
    datetime: str = Field(..., alias='dateTime')
    trace_id: str = Field(..., alias='traceId')


class Bill(HashableBase):
    amount: HashablePlainAmount
    status: BillStatus
    site_id: str = Field(..., alias='siteId')
    id: str = Field(..., alias='billId')
    created_at: datetime = Field(..., alias='creationDateTime')
    expire_at: datetime = Field(..., alias='expirationDateTime')
    pay_url: str = Field(..., alias='payUrl')
    customer: Optional[Customer] = None
    custom_fields: Optional[CustomFields] = Field(None, alias='customFields')

    class Config(BaseConfig):
        extra = Extra.allow
        allow_mutation = True

    @property
    def invoice_uid(self) -> str:
        return self.pay_url[-36:]


class RefundedBill(HashableBase):
    """object: RefundedBill"""

    amount: PlainAmount
    datetime: datetime
    refund_id: str = Field(..., alias='refundId')
    status: str

    def __str__(self) -> str:
        return f'№{self.refund_id} {self.status} {self.amount} {self.datetime}'


class BillWebhookPayload(Bill):
    pay_url: None = Field(None, exclude=True)  # type: ignore


class BillWebhook(HashableBase):
    version: str = Field(..., alias='version')
    bill: BillWebhookPayload = Field(..., alias='bill')

    def __repr__(self) -> str:
        return f'#{self.bill.id} {self.bill.amount} {self.bill.status} '

    def verify_signature(self, sha256_signature: str, secret_p2p_key: str) -> None:
        webhook_key = base64.b64decode(bytes(secret_p2p_key, 'utf-8'))
        bill = self.bill

        invoice_params = f'{bill.amount.currency}|{bill.amount.value}|{bill.id}|{bill.site_id}|{bill.status.value}'
        generated_signature = hmac.new(
            webhook_key, invoice_params.encode('utf-8'), hashlib.sha256
        ).hexdigest()

        if generated_signature != sha256_signature:
            raise WebhookSignatureUnverifiedError()


class PairOfP2PKeys(HashableBase):
    public_key: str = Field(..., alias='PublicKey')
    secret_key: str = Field(..., alias='SecretKey')


class InvoiceStatus(HashableBase):
    invoice_status: str
    is_sms_confirm: bool
    pay_results: Dict[Any, Any] = Field(..., alias='WALLET_ACCEPT_PAY_RESULT')


__all__ = (
    'Bill',
    'BillError',
    'RefundedBill',
    'BillWebhook',
    'PairOfP2PKeys',
    'InvoiceStatus',
    'BillStatus',
    'Customer',
)



from json import JSONDecodeError
from typing import Any, ClassVar

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request, ReturningType
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.exceptions import QiwiAPIError

try:
    from orjson import JSONDecodeError as OrjsonDecodeError
except ImportError:
    from json import JSONDecodeError as OrjsonDecodeError  # type: ignore


class RevealCardID(QiwiAPIMethod[str]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://qiwi.com/card/detect.action'

    card_number: str = Field(..., alias='cardNumber')

    @classmethod
    def on_json_parse(cls, response: HTTPResponse) -> str:
        return response.json()['message']

    def build_request(self, **url_format_kw: Any) -> 'Request':
        r = super().build_request(**url_format_kw)
        r.headers['Content-Type'] = 'application/x-www-form-urlencoded'
        return r



import abc
import logging
from typing import TYPE_CHECKING as MYPY
from typing import Any, Generic, Type, TypeVar

from aiohttp import web
from aiohttp.web_request import Request

from glQiwiApi.core.event_fetching.dispatcher import BaseDispatcher
from glQiwiApi.core.event_fetching.webhooks.dto.errors import WebhookAPIError
from glQiwiApi.core.event_fetching.webhooks.services.collision_detector import (
    AbstractCollisionDetector,
    UnexpectedCollision,
)
from glQiwiApi.utils.compat import json

if MYPY:
    from glQiwiApi.types.base import HashableBase  # noqa

Event = TypeVar('Event', bound='HashableBase')

logger = logging.getLogger('glQiwiApi.webhooks.base')


class BaseWebhookView(web.View, Generic[Event]):
    """
    Generic webhook view for processing events
    You can make your own view and than use it in code, just inheriting this base class

    """

    def __init__(
        self,
        request: Request,
        dispatcher: BaseDispatcher,
        collision_detector: AbstractCollisionDetector[Any],
        event_cls: Type[Event],
        encryption_key: str,
    ) -> None:
        super().__init__(request)
        self._dispatcher = dispatcher
        self._collision_detector = collision_detector
        self._event_cls = event_cls
        self._encryption_key = encryption_key

    @abc.abstractmethod
    async def ok_response(self) -> web.Response:
        pass

    async def get(self) -> web.Response:
        return web.Response(text='')

    async def post(self) -> web.Response:
        event = await self._parse_raw_request()

        try:
            self._collision_detector.remember_processed_object(event)
        except UnexpectedCollision:
            logger.debug('Detect collision on event %s', event)
            return await self.ok_response()

        self._validate_event_signature(event)
        await self.process_event(event)
        return await self.ok_response()

    async def _parse_raw_request(self) -> Event:
        """Parse raw update and return pydantic model"""
        try:
            data = await self.request.json(loads=json.loads)
            if isinstance(data, str):
                return self._event_cls.parse_raw(data)
            elif isinstance(data, dict):  # pragma: no cover
                return self._event_cls.parse_obj(data)
            else:
                raise ValueError()
        except ValueError:
            raise web.HTTPBadRequest(
                text=WebhookAPIError(status='Validation error').json(),
                content_type='application/json',
            )

    @abc.abstractmethod
    def _validate_event_signature(self, update: Event) -> None:
        raise NotImplementedError

    async def process_event(self, event: Event) -> None:
        await self._dispatcher.process_event(event)



from typing import ClassVar, Union

from pydantic import Field

from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import TransactionType
from glQiwiApi.types.arbitrary import BinaryIOInput, File


class GetReceipt(QiwiAPIMethod[File]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/payment-history/v1/transactions/{transaction_id}/cheque/file'
    http_method: ClassVar[str] = 'GET'

    transaction_id: Union[str, int] = Field(..., path_runtime_value=True)
    transaction_type: TransactionType = Field(..., alias='type')
    file_format: str = Field(alias='format', default='PDF')

    class Config:
        use_enum_values = True

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> File:  # type: ignore
        return File(BinaryIOInput.from_bytes(response.body))



from typing import Any, ClassVar, Dict, Union

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import RefundedBill
from glQiwiApi.types.amount import PlainAmount


class RefundBill(QiwiAPIMethod[RefundedBill]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}/refunds/{refund_id}'

    bill_id: str = Field(..., path_runtime_value=True)
    refund_id: str = Field(..., path_runtime_value=True)

    json_bill_data: Union[PlainAmount, Dict[str, Union[str, int]]]

    def build_request(self, **url_format_kw: Any) -> 'Request':
        json_payload = self.json_bill_data
        if isinstance(self.json_bill_data, PlainAmount):
            json_payload = self.json_bill_data.json(encoder=self.Config.json_dumps)  # type: ignore

        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            json_payload=json_payload,
            http_method=self.http_method,
        )



import typing as t

from aiohttp import web

from glQiwiApi.core.event_fetching.dispatcher import BaseDispatcher
from glQiwiApi.core.event_fetching.webhooks.config import WebhookConfig
from glQiwiApi.core.event_fetching.webhooks.middlewares.ip import ip_filter_middleware
from glQiwiApi.core.event_fetching.webhooks.services.collision_detector import (
    HashBasedCollisionDetector,
)
from glQiwiApi.core.event_fetching.webhooks.services.security.ip import IPFilter
from glQiwiApi.core.event_fetching.webhooks.utils import inject_dependencies
from glQiwiApi.core.event_fetching.webhooks.views.bill_view import QiwiBillWebhookView
from glQiwiApi.core.event_fetching.webhooks.views.transaction_view import (
    QiwiTransactionWebhookView,
)
from glQiwiApi.qiwi.clients.p2p.types import BillWebhook
from glQiwiApi.qiwi.clients.wallet.types.webhooks import TransactionWebhook


def configure_app(
    dispatcher: BaseDispatcher, app: web.Application, webhook_config: WebhookConfig
) -> web.Application:
    """
    Entirely configures the web app for webhooks.

    :param dispatcher: dispatcher, which processing events
    :param app: aiohttp.web.Application
    :param webhook_config:
    """

    generic_dependencies: t.Dict[str, t.Any] = {
        'dispatcher': dispatcher,
        'collision_detector': HashBasedCollisionDetector(),
    }

    app.router.add_view(
        handler=inject_dependencies(
            QiwiBillWebhookView,
            {
                'event_cls': BillWebhook,
                'encryption_key': webhook_config.encryption.secret_p2p_key,
                **generic_dependencies,
            },
        ),
        name=webhook_config.routes.p2p_view_route_name,
        path=webhook_config.routes.p2p_path,
    )

    app.router.add_view(
        handler=inject_dependencies(
            QiwiTransactionWebhookView,
            {
                'event_cls': TransactionWebhook,
                'encryption_key': webhook_config.encryption.base64_encryption_key,
                **generic_dependencies,
            },
        ),
        name=webhook_config.routes.standard_qiwi_route_name,
        path=webhook_config.routes.standard_qiwi_hook_path,
    )

    if webhook_config.security.check_ip:
        app.middlewares.append(ip_filter_middleware(IPFilter.default()))

    return app



from typing import ClassVar, Optional, Union

from pydantic import Field

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.yoo_money.types import RequestPaymentResponse


class RequestPayment(APIMethod[RequestPaymentResponse]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/request-payment'

    to_account: str = Field(..., alias='to')
    amount: Union[int, float] = Field(..., alias='amount_due')
    pattern_id: str = 'p2p'
    comment_for_history: Optional[str] = Field(None, alias='comment')
    comment_for_receiver: Optional[str] = Field(None, alias='message')
    protect: bool = Field(False, alias='codepro')
    expire_period: int = 1



from datetime import datetime
from typing import Dict, List, Optional, Union
from urllib.parse import urljoin, urlparse

from glQiwiApi.core.abc.base_api_client import BaseAPIClient, RequestServiceFactoryType
from glQiwiApi.core.request_service import RequestService, RequestServiceProto
from glQiwiApi.core.session import AiohttpSessionHolder
from glQiwiApi.qiwi.clients.p2p.methods.create_p2p_bill import CreateP2PBill
from glQiwiApi.qiwi.clients.p2p.methods.create_p2p_key_pair import CreateP2PKeyPair
from glQiwiApi.qiwi.clients.p2p.methods.get_bill_by_id import GetBillByID
from glQiwiApi.qiwi.clients.p2p.methods.refund_bill import RefundBill
from glQiwiApi.qiwi.clients.p2p.methods.reject_p2p_bill import RejectP2PBill
from glQiwiApi.qiwi.clients.p2p.types import Bill, Customer, PairOfP2PKeys, RefundedBill
from glQiwiApi.types.amount import PlainAmount
from glQiwiApi.utils.compat import remove_suffix
from glQiwiApi.utils.deprecated import warn_deprecated
from glQiwiApi.utils.validators import String


class NoShimUrlWasProvidedError(Exception):
    pass


DEPRECATED_SHIM_URL_SUFFIXES = ('{0}', '{}')


def _parse_shim_url(shim_url: Optional[str]) -> Optional[str]:
    if shim_url is None:
        return None

    shim_url = urlparse(shim_url)

    if any(shim_url.path.endswith(suffix) for suffix in DEPRECATED_SHIM_URL_SUFFIXES):
        new_format_url_example = str(shim_url.geturl())
        for suffix in DEPRECATED_SHIM_URL_SUFFIXES:
            new_format_url_example = remove_suffix(new_format_url_example, suffix)

        warn_deprecated(
            'Old-style urls that were used like format-like strings are deprecated '
            f'use plain path like this - {new_format_url_example} instead.'
        )

        shim_url = urlparse(new_format_url_example)

    return str(shim_url.geturl())


class QiwiP2PClient(BaseAPIClient):
    _api_access_token = String(optional=False)

    def __init__(
        self,
        secret_p2p: str,
        request_service_factory: Optional[RequestServiceFactoryType] = None,
        shim_server_url: Optional[str] = None,
    ) -> None:
        """
        :param secret_p2p: QIWI P2P secret key received from https://p2p.qiwi.com/
        :param shim_server_url:
        """
        super().__init__(request_service_factory)
        self._api_access_token = secret_p2p
        self._shim_server_url = _parse_shim_url(shim_server_url)

    async def _create_request_service(self) -> RequestServiceProto:
        return RequestService(
            session_holder=AiohttpSessionHolder(
                headers={
                    'Authorization': f'Bearer {self._api_access_token}',
                    'Accept': 'application/json',
                    'Content-Type': 'application/json',
                }
            )
        )

    async def get_bill_by_id(self, bill_id: str) -> Bill:
        """
        Method for checking the status of a p2p transaction.\n
        Possible transaction types: \n
        WAITING	    Bill is waiting for pay	\n
        PAID	    Bill was paid	\n
        REJECTED	Bill was rejected\n
        EXPIRED	The bill has expired. Invoice not paid\n
        Docs: https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id:
        :return: status of bill
        """
        return await self._request_service.execute_api_method(GetBillByID(bill_id=bill_id))

    async def check_if_bill_was_paid(self, bill: Bill) -> bool:
        bill_status = await self.get_bill_status(bill.id)
        return bill_status == 'PAID'

    async def get_bill_status(self, bill_id: str) -> str:
        """
        Method for checking the status of a p2p transaction.\n
        Possible transaction types: \n
        WAITING	    Bill is waiting for pay	\n
        PAID	    Bill was paid	\n
        REJECTED	Bill was rejected\n
        EXPIRED	The bill has expired. Invoice not paid\n
        Docs: https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id:
        :return: status of bill
        """
        return (await self.get_bill_by_id(bill_id)).status.value

    async def create_p2p_bill(
        self,
        amount: Union[int, float, str],
        bill_id: Optional[str] = None,
        comment: Optional[str] = None,
        expire_at: Optional[datetime] = None,
        theme_code: Optional[str] = None,
        pay_source_filter: Optional[List[str]] = None,
        customer: Optional[Customer] = None,
    ) -> Bill:
        """
        It is the reliable method for integration.
        Parameters are sent by means of server2server requests with authorization.
        Method allows you to issue an invoice, successful
         response contains payUrl link to redirect client on Payment Form.
        Possible values of pay_source_filter:
          - 'qw'
          - 'card'
          - 'mobile'

        :param amount: amount of payment
        :param bill_id: unique transaction number, if not transmitted,
         generated automatically,
        :param expire_at: the date until which the invoice will be available for payment.
        :param comment:
        :param theme_code:
        :param pay_source_filter: When you open the form, the following will be displayed
         only the translation methods specified in this parameter
        :param customer:
        """
        return await self._request_service.execute_api_method(
            CreateP2PBill(
                bill_id=bill_id,
                expire_at=expire_at,
                amount=amount,
                comment=comment,
                theme_code=theme_code,
                pay_source_filter=pay_source_filter,
                customer=customer,
            )
        )

    async def reject_bill(self, bill: Bill) -> Bill:
        return await self.reject_p2p_bill(bill.id)

    async def reject_p2p_bill(self, bill_id: str) -> Bill:
        """Use this method to cancel unpaid invoice."""
        return await self._request_service.execute_api_method(RejectP2PBill(bill_id=bill_id))

    async def create_pair_of_p2p_keys(
        self, key_pair_name: str, server_notification_url: Optional[str] = None
    ) -> PairOfP2PKeys:
        """
        Creates a new pair of P2P keys to interact with P2P QIWI API

        :param key_pair_name: P2P token pair name
        :param server_notification_url: endpoint for webhooks
        """
        return await self._request_service.execute_api_method(
            CreateP2PKeyPair(
                key_pair_name=key_pair_name, server_notification_url=server_notification_url
            )
        )

    async def refund_bill(
        self,
        bill_id: Union[str, int],
        refund_id: Union[str, int],
        json_bill_data: Union[PlainAmount, Dict[str, Union[str, int]]],
    ) -> RefundedBill:
        """
        The method allows you to make a refund on a paid invoice.
         in the JSON body of the request for the json_bill_data parameter:
         amount.value - refund amount.
         amount.currency - return currency.
        Can be a dictionary or an PlainAmount object
         Dictionary example: {
        "amount": {
            "currency": "RUB",
            "value": 1
            }
        }

        :param bill_id: unique account identifier in the merchant's system
        :param refund_id: unique identifier of the refund in the merchant's system.
        :param json_bill_data:
        :return: RefundBill object
        """
        return await self._request_service.execute_api_method(
            RefundBill(bill_id=bill_id, refund_id=refund_id, json_bill_data=json_bill_data)
        )

    def create_shim_url(self, bill_or_invoice_uid: Union[Bill, str]) -> str:
        if self._shim_server_url is None:
            raise NoShimUrlWasProvidedError(
                "QiwiP2PClient has no shim endpoint -> can't create shim endpoint for bill"
            )

        invoice_uid = bill_or_invoice_uid
        if isinstance(bill_or_invoice_uid, Bill):
            invoice_uid = bill_or_invoice_uid.invoice_uid

        return urljoin(self._shim_server_url, invoice_uid)



from __future__ import annotations

import asyncio
import inspect
import pathlib
from typing import Any, BinaryIO, Union

from glQiwiApi.types.arbitrary.inputs import AbstractInput
from glQiwiApi.utils.compat import aiofiles

CHUNK_SIZE = 65536

StrOrBytesPath = Union[str, bytes, pathlib.Path]  # stable

_OpenFile = Union[StrOrBytesPath, int]


class File:
    def __init__(self, input: AbstractInput[Any]) -> None:
        self._input = input

    def get_filename(self) -> str:
        return self._input.get_filename()

    def get_underlying_file_descriptor(self) -> BinaryIO:
        return self._input.get_file()

    def get_path(self) -> str:
        return self._input.get_path()

    def save(self, path: StrOrBytesPath, chunk_size: int = CHUNK_SIZE) -> None:
        file_descriptor = self.get_underlying_file_descriptor()
        with open(path, 'wb') as fp:
            while True:
                data = file_descriptor.read(chunk_size)
                if not data:
                    break
                fp.write(data)
            fp.flush()

        if file_descriptor.seekable():
            file_descriptor.seek(0)

    async def save_asynchronously(
        self, path: StrOrBytesPath, chunk_size: int = CHUNK_SIZE
    ) -> None:
        file_descriptor = self.get_underlying_file_descriptor()
        async with aiofiles.open(path, 'wb') as fp:
            while True:
                data = file_descriptor.read(chunk_size)
                if not data:
                    break
                await fp.write(data)
            await fp.flush()

        if file_descriptor.seekable():
            file_descriptor.seek(0)

    def __str__(self) -> str:
        try:
            return self.get_filename()
        except TypeError:
            return '<File(binary stream underlies)>'

    def __del__(self) -> None:
        if not hasattr(self, '_input'):
            return

        if inspect.iscoroutinefunction(self._input.close()):  # type: ignore  # noqa
            return asyncio.ensure_future(self._input.close())  # type: ignore # noqa
        self._input.close()

    __repr__ = __str__



from typing import Any, ClassVar, Dict

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod


class DeleteWebhook(QiwiAPIMethod[Dict[Any, Any]]):
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-notifier/v1/hooks/{hook_id}'
    http_method: ClassVar[str] = 'DELETE'

    hook_id: str

    def build_request(self, **url_format_kw: Any) -> Request:
        return super().build_request(**url_format_kw, hook_id=self.hook_id)



from datetime import datetime
from typing import ClassVar

from pydantic import Field

from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.types.arbitrary import BinaryIOInput, File


class GetQiwiMasterStatement(QiwiAPIMethod[File]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/payment-history/v1/persons/{phone_number}/cards/{card_id}/statement'
    http_method: ClassVar[str] = 'GET'

    card_id: str = Field(..., path_runtime_value=True)
    from_date: datetime = Field(..., alias='from')
    till_date: datetime = Field(..., alias='till')

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> File:  # type: ignore
        return File(BinaryIOInput.from_bytes(response.body))



from typing import Any, ClassVar

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import UserProfile


class GetUserProfile(QiwiAPIMethod[UserProfile]):
    url: ClassVar[str] = 'https://edge.qiwi.com/person-profile/v1/profile/current'
    http_method: ClassVar[str] = 'GET'

    include_auth_info: bool = Field(True, alias='authInfoEnabled')
    include_contract_info: bool = Field(True, alias='contractInfoEnabled')
    include_user_info: bool = Field(True, alias='userInfoEnabled')

    def build_request(self, **url_format_kw: Any) -> 'Request':
        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
            params={
                'authInfoEnabled': str(self.include_auth_info),
                'contractInfoEnabled': str(self.include_contract_info),
                'userInfoEnabled': str(self.include_user_info),
            },
        )



from datetime import datetime
from typing import Any, ClassVar, Iterable, Optional, Union

from pydantic import conint

from glQiwiApi.core.abc.api_method import APIMethod, Request
from glQiwiApi.utils.date_conversion import datetime_to_utc_in_iso_format
from glQiwiApi.utils.payload import filter_dictionary_none_values
from glQiwiApi.yoo_money.types import OperationHistory

MAX_HISTORY_LIMIT = 100


class OperationHistoryMethod(APIMethod[OperationHistory]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://yoomoney.ru/api/operation-history'

    operation_types: Optional[Iterable[str]] = None
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    start_record: Optional[int] = None
    records: conint(le=MAX_HISTORY_LIMIT, strict=True, gt=0) = 30
    label: Optional[Union[str, int]] = None
    in_detail: bool = False

    def build_request(self, **url_format_kw: Any) -> 'Request':
        payload = {
            'records': self.records,
            'label': self.label,
            'start_record': self.start_record,
            'details': str(self.in_detail).lower(),
        }

        if self.operation_types is not None:
            payload['type'] = ' '.join([op_type.lower() for op_type in self.operation_types])
        if self.start_date:
            payload['from'] = datetime_to_utc_in_iso_format(self.start_date)
        if self.end_date:
            payload['till'] = datetime_to_utc_in_iso_format(self.end_date)

        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
            data=filter_dictionary_none_values(payload),
        )



from typing import ClassVar, Optional

from glQiwiApi.core.abc.api_method import APIMethod, ReturningType
from glQiwiApi.core.session.holder import HTTPResponse


class GetAccessToken(APIMethod[str]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/oauth/token'

    code: str
    client_id: str
    grant_type: str = 'authorization_code'
    redirect_uri: str = 'https://example.com'
    client_secret: Optional[str] = None

    @classmethod
    def on_json_parse(cls, response: HTTPResponse) -> str:
        return response.json()['access_token']



from typing import Any, ClassVar

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Transaction, TransactionType


class GetTransactionInfo(QiwiAPIMethod[Transaction]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-history/v1/transactions/{transaction_id}'

    transaction_id: int
    transaction_type: TransactionType = Field(..., alias='type')

    def build_request(self, **url_format_kw: Any) -> 'Request':
        return super().build_request(**url_format_kw, transaction_id=self.transaction_id)

    class Config:
        use_enum_values = True



import logging

from aiohttp import web

from glQiwiApi.core.event_fetching.webhooks.dto.errors import WebhookAPIError
from glQiwiApi.core.event_fetching.webhooks.views.base import BaseWebhookView
from glQiwiApi.qiwi.clients.wallet.types.webhooks import TransactionWebhook
from glQiwiApi.types.exceptions import WebhookSignatureUnverifiedError

logger = logging.getLogger('glQiwiApi.webhooks.transaction')


class QiwiTransactionWebhookView(BaseWebhookView[TransactionWebhook]):
    def _validate_event_signature(self, update: TransactionWebhook) -> None:
        if update.is_experimental:  # pragma: no cover
            return None

        logger.debug('Current encryption key is %s', self._encryption_key)

        try:
            update.verify_signature(self._encryption_key)
        except WebhookSignatureUnverifiedError:
            logger.debug(
                'Request has being blocked due to invalid signature of json request payload.'
            )
            raise web.HTTPBadRequest(
                text=WebhookAPIError(status='Invalid hash of transaction.').json(),
                content_type='application/json',
            )

    async def ok_response(self) -> web.Response:
        return web.Response(text='ok')



from typing import ClassVar

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.nickname import NickName


class GetNickName(QiwiAPIMethod[NickName]):
    url: ClassVar[str] = 'https://edge.qiwi.com/qw-nicknames/v1/persons/{phone_number}/nickname'
    http_method: ClassVar[str] = 'GET'



from typing import ClassVar, List

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Balance
from glQiwiApi.qiwi.clients.wallet.types.balance import AvailableBalance


class GetAvailableBalances(QiwiAPIMethod[List[AvailableBalance]]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/funding-sources/v2/persons/{phone_number}/accounts/offer'



from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional

from pydantic import Field, conint, root_validator

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import History, Source, TransactionType
from glQiwiApi.utils.date_conversion import datetime_to_iso8601_with_moscow_timezone

MAX_HISTORY_LIMIT = 50


class GetHistory(QiwiAPIMethod[History]):
    http_method: ClassVar[str] = 'GET'
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-history/v2/persons/{phone_number}/payments'

    rows: conint(le=MAX_HISTORY_LIMIT, strict=True, gt=0) = MAX_HISTORY_LIMIT
    transaction_type: TransactionType = Field(TransactionType.ALL, alias='operation')
    sources: Optional[List[Source]] = Field(None)
    start_date: Optional[datetime] = Field(None, alias='startDate')
    end_date: Optional[datetime] = Field(None, alias='endDate')
    next_txn_date: Optional[datetime] = Field(None, alias='nextTxnDate')
    next_txn_id: Optional[int] = Field(None, alias='nextTxnId')

    class Config:
        use_enum_values = True

    @root_validator()
    def check_if_start_date_is_not_greater_than_end_date(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        start_date: Optional[datetime] = values.get('start_date')
        end_date: Optional[datetime] = values.get('end_date')

        both_are_not_transmitted = start_date is None and end_date is None

        if both_are_not_transmitted:
            return values

        only_start_date_transmitted = start_date is not None and end_date is None
        only_end_date_transmitted = start_date is None and end_date is not None

        if only_start_date_transmitted or only_end_date_transmitted:
            raise ValueError(
                'You must transmit both start_date and end_date parameters to get history rather than one'
            )

        if (end_date - start_date).total_seconds() <= 0:  # type: ignore
            raise ValueError('start_date cannot be bigger than end_date')

        return values

    @root_validator()
    def check_next_txn_date_and_next_txn_id(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        next_txn_id: Optional[int] = values.get('next_txn_id')
        next_txn_date: Optional[datetime] = values.get('next_txn_date')

        both_are_not_transmitted = next_txn_id is None and next_txn_date is None

        if both_are_not_transmitted:
            return values

        only_next_txn_date_transmitted = next_txn_date is not None and next_txn_id is None
        only_next_txn_id_transmitted = next_txn_date is None and next_txn_id is not None

        if only_next_txn_date_transmitted or only_next_txn_id_transmitted:
            raise ValueError(
                'You must transmit both next_txn_id and next_txn_date '
                'parameters to get history rather than one'
            )

        return values

    def build_request(self, **url_format_kw: Any) -> Request:
        data = self.dict(exclude={'sources'}, exclude_none=True, by_alias=True)

        if data.get('startDate') and data.get('endDate'):
            data['startDate'] = datetime_to_iso8601_with_moscow_timezone(data['startDate'])
            data['endDate'] = datetime_to_iso8601_with_moscow_timezone(data['endDate'])

        if data.get('nextTxnDate'):
            data['nextTxnDate'] = datetime_to_iso8601_with_moscow_timezone(data['nextTxnDate'])

        if self.sources is None:
            return Request(endpoint=self.url.format(**url_format_kw), params=data)

        for index, source in enumerate(self.sources, start=1):
            data.update({f'source[{index}]': source})

        return Request(endpoint=self.url.format(**url_format_kw), params=data)



import abc
import re
from abc import ABC
from typing import (
    Any,
    Callable,
    Generic,
    Match,
    Optional,
    Pattern,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from glQiwiApi.utils.compat import Literal


class ValidationError(Exception):
    pass


PHONE_NUMBER_PATTERN: Pattern[str] = re.compile(
    r'^[+]?[(]?[0-9]{3}[)]?[-\s.]?[0-9]{3}[-\s.]?[0-9]{4,6}$'
)

_FieldType = TypeVar('_FieldType', bound=Any)


class Field(ABC, Generic[_FieldType]):
    def __init__(self, *validators: Callable[[_FieldType], None]):
        self._validators = validators

    def __set_name__(self, owner: Type[Any], name: str) -> None:
        self.private_name = '_' + name

    def __get__(self, obj: Any, objtype: Optional[Type[Any]] = None) -> _FieldType:
        return cast(_FieldType, getattr(obj, self.private_name))

    def __set__(self, obj: Any, value: _FieldType) -> None:
        self.validate(value)
        setattr(obj, self.private_name, value)

    def validate(self, value: _FieldType) -> None:
        for validator in self._validators:
            validator(value)


class AbstractValidator(ABC):
    def __init__(self, optional: bool = False):
        self._optional = optional

    def validate(self, value: Any) -> None:
        if self._optional is True and value is None:
            return
        self._validate(value)

    @abc.abstractmethod
    def _validate(self, value: Any) -> None:
        pass

    def __call__(self, value: Any) -> None:
        self.validate(value)


class StringValidator(AbstractValidator):
    def __init__(
        self,
        *,
        minsize: Optional[int] = None,
        maxsize: Optional[int] = None,
        predicate: Optional[Callable[[Any], bool]] = None,
        optional: bool = False,
    ):
        AbstractValidator.__init__(self, optional=optional)
        self.minsize = minsize
        self.maxsize = maxsize
        self.predicate = predicate
        self._optional = optional

    def _validate(self, value: Any) -> None:
        if not isinstance(value, str):
            raise ValidationError(f'Expected {value!r} to be an str')
        if self.minsize is not None and len(value) < self.minsize:
            raise ValidationError(f'Expected {value!r} to be no smaller than {self.minsize!r}')
        if self.maxsize is not None and len(value) > self.maxsize:
            raise ValidationError(f'Expected {value!r} to be no bigger than {self.maxsize!r}')
        if self.predicate is not None and not self.predicate(value):
            raise ValidationError(f'Expected {self.predicate} to be true for {value!r}')


class PhoneNumberValidator(StringValidator):
    def _validate(self, value: Optional[str]) -> None:
        StringValidator._validate(self, value)
        phone_number_match: Optional[Match[Any]] = re.fullmatch(PHONE_NUMBER_PATTERN, value)
        if not phone_number_match:
            raise ValidationError(
                'Failed to verify parameter `phone_number` by regex. '
                'Please, enter the correct phone number.'
            )
        if not value.startswith('+'):  # type: ignore
            raise ValidationError(f'Expected {value!r} starts with + sign')


class IntegerValidator(AbstractValidator):
    def _validate(self, value: _FieldType) -> None:
        if not isinstance(value, int):
            raise ValidationError(f'Expected {value!r} to be an integer')


@overload
def String(optional: Literal[True], **options: Any) -> Field[Optional[str]]:
    ...


@overload
def String(optional: Literal[False], **options: Any) -> Field[str]:
    ...


def String(optional: bool = False, **options: Any) -> Union[Field[Optional[str]], Field[str]]:
    return Field(StringValidator(optional=optional, **options))


def PhoneNumber(**options: Any) -> Field[Optional[str]]:
    return Field(PhoneNumberValidator(**options))


@overload
def Integer(optional: Literal[True]) -> Field[Optional[int]]:
    ...


@overload
def Integer(optional: Literal[False]) -> Field[int]:
    ...


def Integer(optional: bool = False) -> Union[Field[Optional[int]], Field[int]]:
    return Field(IntegerValidator(optional=optional))



from typing import Any, ClassVar

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import WebhookInfo


class RegisterWebhook(QiwiAPIMethod[WebhookInfo]):
    url: ClassVar[str] = 'https://edge.qiwi.com/payment-notifier/v1/hooks'
    http_method: ClassVar[str] = 'PUT'

    webhook_url: str = Field(..., alias='param')
    txn_type: int = Field(..., alias='txnType')

    hook_type: int = Field(default=1, alias='hookType')

    def build_request(self, **url_format_kw: Any) -> 'Request':
        return Request(
            endpoint=self.url.format(**url_format_kw),
            params=self.dict(by_alias=True),
            http_method=self.http_method,
        )



from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod


class SetDefaultBalance(QiwiAPIMethod[Dict[Any, Any]]):
    http_method: ClassVar[str] = 'PATCH'
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/funding-sources/v2/persons/{phone_number}/accounts/{account_alias}'

    account_alias: str = Field(..., path_runtime_value=True)



from pydantic import BaseConfig, BaseModel

from glQiwiApi.utils.compat import json


class Base(BaseModel):
    class Config(BaseConfig):
        json_dumps = json.dumps  # type: ignore
        json_loads = json.loads
        orm_mode = True


class HashableBase(Base):
    class Config(BaseConfig):
        allow_mutation = False
        frozen = True



from typing import Any, Dict, Optional, Union

from pydantic import Field, validator

from glQiwiApi.types.amount import AmountWithCurrency, CurrencyModel
from glQiwiApi.types.base import HashableBase
from glQiwiApi.utils.currency_util import Currency


class AvailableBalance(HashableBase):
    alias: str
    currency: Union[str, CurrencyModel]


class Balance(HashableBase):
    """object: Balance"""

    alias: str
    title: str
    fs_alias: str = Field(alias='fsAlias')
    bank_alias: str = Field(alias='bankAlias')
    has_balance: bool = Field(alias='hasBalance')
    balance: Optional[AmountWithCurrency] = None
    currency: CurrencyModel
    account_type: Optional[Dict[str, Any]] = Field(None, alias='type')
    is_default_account: bool = Field(alias='defaultAccount')

    @validator('currency', pre=True)
    def humanize_pay_currency(cls, v):  # type: ignore
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))



from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import InvoiceStatus


class PayInvoice(QiwiAPIMethod[InvoiceStatus]):
    url: ClassVar[str] = 'https://edge.qiwi.com/checkout-api/invoice/pay/wallet'
    http_method: ClassVar[str] = 'POST'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'invoice_uid': RuntimeValue(),
        'currency': RuntimeValue(),
    }

    invoice_uid: str = Field(..., schema_path='invoice_uid')
    currency: str = Field(..., schema_path='currency')



from typing import ClassVar

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.yoo_money.types import Payment


class ProcessPayment(APIMethod[Payment]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/process-payment'

    request_id: str
    money_source: str = 'wallet'



from typing import ClassVar

from pydantic import Field

from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.qiwi_master import OrderDetails


class ConfirmQiwiMasterPurchaseOrder(QiwiAPIMethod[OrderDetails]):
    url: ClassVar[
        str
    ] = 'https://edge.qiwi.com/cards/v2/persons/{phone_number}/orders/{order_id}/submit'
    http_method: ClassVar[str] = 'PUT'

    order_id: str = Field(..., path_runtime_value=True)



from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

from glQiwiApi.qiwi.clients.p2p.client import QiwiP2PClient
from glQiwiApi.qiwi.clients.p2p.types import (
    Bill,
    Customer,
    InvoiceStatus,
    PairOfP2PKeys,
    RefundedBill,
)
from glQiwiApi.qiwi.clients.wallet.client import AmountType, QiwiWallet
from glQiwiApi.qiwi.clients.wallet.methods.get_limits import ALL_LIMIT_TYPES
from glQiwiApi.qiwi.clients.wallet.methods.history import MAX_HISTORY_LIMIT
from glQiwiApi.qiwi.clients.wallet.types import (
    Balance,
    Card,
    Commission,
    CrossRate,
    History,
    Identification,
    Limit,
    OrderDetails,
    PaymentDetails,
    PaymentInfo,
    PaymentMethod,
    Restriction,
    Source,
    Statistic,
    Transaction,
    TransactionType,
    UserProfile,
    WebhookInfo,
)
from glQiwiApi.qiwi.clients.wallet.types.balance import AvailableBalance
from glQiwiApi.qiwi.clients.wallet.types.mobile_operator import MobileOperator
from glQiwiApi.qiwi.clients.wallet.types.nickname import NickName
from glQiwiApi.types.amount import AmountWithCurrency, PlainAmount
from glQiwiApi.types.arbitrary import File
from glQiwiApi.utils.deprecated import deprecated

if TYPE_CHECKING:
    from glQiwiApi.ext.webhook_url import WebhookURL

_T = TypeVar('_T', bound='QiwiWrapper')


class QiwiWrapper:
    """For backward compatibility with glQiwiApi <= 1.1.4"""

    def __init__(
        self,
        api_access_token: Optional[str] = None,
        phone_number: Optional[str] = None,
        secret_p2p: Optional[str] = None,
        shim_server_url: Optional[str] = None,
    ) -> None:
        self._qiwi_wallet = QiwiWallet(api_access_token or '', phone_number)
        self._p2p_client = QiwiP2PClient(secret_p2p or '', shim_server_url=shim_server_url)

    async def get_balance(self, *, account_number: int = 1) -> AmountWithCurrency:
        return await self._qiwi_wallet.get_balance(account_number=account_number)

    async def get_nickname(self) -> NickName:
        return await self._qiwi_wallet.get_nickname()

    async def transactions(
        self,
        rows: int = 50,
        operation: TransactionType = TransactionType.ALL,
        sources: Optional[List[Source]] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        next_txn_date: Optional[datetime] = None,
        next_txn_id: Optional[int] = None,
    ) -> History:
        """
        Method for receiving transactions on the account
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param rows: number of transactions you want to receive
        :param operation: The type of operations in the report for selection.
        :param sources: List of payment sources, for filter
        :param start_date: The starting date for searching for payments.
                            Used only in conjunction with end_date.
        :param end_date: the end date of the search for payments.
                            Used only in conjunction with start_date.
        :param next_txn_id:
        :param next_txn_date:
        """
        return await self._qiwi_wallet.history(
            rows,
            operation,
            sources,
            start_date,
            end_date,
            next_txn_id=next_txn_id,
            next_txn_date=next_txn_date,
        )

    async def transaction_info(
        self, transaction_id: Union[str, int], transaction_type: TransactionType
    ) -> Transaction:
        """
        Method for obtaining complete information about a transaction

        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#txn_info

        :param transaction_id:
        :param transaction_type: only IN or OUT
        :return: Transaction object
        """
        return await self._qiwi_wallet.get_transaction_info(transaction_id, transaction_type)

    async def get_restriction(self) -> List[Restriction]:
        """
        Method to check limits on your qiwi wallet
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#restrictions

        :return: List where the dictionary is located with restrictions,
         if there are no restrictions, it returns an empty list
        """
        return await self._qiwi_wallet.get_restrictions()

    async def get_identification(self) -> Identification:
        """
        This method allows get your wallet identification data
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#ident
        """
        return await self._qiwi_wallet.get_identification()

    async def check_transaction(
        self,
        amount: AmountType,
        transaction_type: TransactionType = TransactionType.IN,
        sender: Optional[str] = None,
        rows_num: int = 50,
        comment: Optional[str] = None,
    ) -> bool:
        """
        [ NON API METHOD ]

        Method for verifying a transaction.
        This method uses self.transactions (rows = rows) "under the hood" to check payment.

        For a little optimization, you can decrease rows by setting it,
        however, this does not guarantee the correct result

        Possible values for the transaction_type parameter:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'


        :param amount: amount of payment
        :param transaction_type: type of payment
        :param sender: number of receiver
        :param rows_num: number of payments to be checked
        :param comment: comment by which the transaction will be verified
        """

        def obsolete_check_transaction(txn: Transaction) -> bool:
            if txn.sum.amount < amount or txn.type != transaction_type.value:
                return False
            if txn.comment == comment and txn.to_account == sender:
                return True
            elif comment and sender:
                return False
            elif txn.to_account == sender:
                return True
            elif sender:
                return False
            elif txn.comment == comment:
                return True
            elif comment:
                return False
            return False

        return await self._qiwi_wallet.check_whether_transaction_exists(
            check_fn=obsolete_check_transaction, rows_num=rows_num
        )

    async def check_whether_transaction_exists(
        self, check_fn: Callable[[Transaction], bool], rows_num: int = MAX_HISTORY_LIMIT
    ) -> bool:
        return await self._qiwi_wallet.check_whether_transaction_exists(
            check_fn=check_fn, rows_num=rows_num
        )

    async def get_limits(self, limit_types: List[str] = ALL_LIMIT_TYPES) -> Dict[str, Limit]:
        """
        Function for getting limits on the qiwi wallet account
        Returns wallet limits as a list,
        if there is no limit for a certain country, then it does not include it in the list
        Detailed documentation:

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#limits
        """
        return await self._qiwi_wallet.get_limits(limit_types)

    async def get_list_of_cards(self) -> List[Card]:
        return await self._qiwi_wallet.get_list_of_cards()

    async def authenticate(
        self,
        birth_date: str,
        first_name: str,
        last_name: str,
        middle_name: str,
        passport: str,
        oms: Optional[str] = None,
        inn: Optional[str] = None,
        snils: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        This request allows you to transfer_money data to identify your QIWI wallet.
        It is allowed to identify no more than 5 wallets per owner

        To identify the wallet, you must transfer_money your full name, passport series number and date of birth.
        If the data has been verified, then the response will display
        your TIN and simplified wallet identification will be installed.
        If the data has not been verified,
        the wallet remains in the "Minimum" status.

        :param birth_date: Date of birth as a format string 1998-02-11
        :param first_name: First name
        :param last_name: Last name
        :param middle_name: Middle name
        :param passport: Series / Number of the passport. Ex: 4400111222
        :param oms:
        :param snils:
        :param inn:
        """
        return await self._qiwi_wallet.authenticate(
            birth_date, first_name, last_name, middle_name, passport, oms, inn, snils
        )

    async def get_receipt(
        self,
        transaction_id: Union[str, int],
        transaction_type: TransactionType,
        file_format: str = 'PDF',
    ) -> File:
        """
        Method for receiving a receipt in byte format or file. \n
        Possible transaction_type values:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'

        :param transaction_id: transaction id, can be obtained by calling the transfer_money method,
         transfer_money_to_card
        :param transaction_type: type of transaction: 'IN', 'OUT', 'QIWI_CARD'
        :param file_format: format of file
        """
        return await self._qiwi_wallet.get_receipt(transaction_id, transaction_type, file_format)

    async def get_account_info(self) -> UserProfile:
        return await self._qiwi_wallet.get_profile()

    async def fetch_statistics(
        self,
        start_date: datetime,
        end_date: datetime,
        operation: TransactionType = TransactionType.ALL,
        sources: Optional[List[str]] = None,
    ) -> Statistic:
        """
        This query is used to get summary statistics
        by the amount of payments for a given period.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param start_date:The start date of the statistics period.
        :param end_date: End date of the statistics period.
        :param operation: The type of operations taken into account when calculating statistics.
         Allowed values:
            ALL - все операции,
            IN - только пополнения,
            OUT - только платежи,
            QIWI_CARD - только платежи по картам QIWI (QVC, QVP).
            По умолчанию ALL.
        :param sources: The sources of payments
            QW_RUB - рублевый счет кошелька,
            QW_USD - счет кошелька в долларах,
            QW_EUR - счет кошелька в евро,
            CARD - привязанные и непривязанные к кошельку банковские карты,
            MK - счет мобильного оператора. Если не указан,
            учитываются все источники платежа.
        """
        return await self._qiwi_wallet.fetch_statistics(start_date, end_date, operation, sources)

    async def list_of_balances(self) -> List[Balance]:
        """
        The request gets the current account balances of your QIWI Wallet.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#balances_list

        """
        return await self._qiwi_wallet.get_list_of_balances()

    async def create_new_balance(self, currency_alias: str) -> Dict[str, bool]:
        """
        The request creates a new account and balance in your QIWI Wallet

        :param currency_alias: New account alias
        """
        return await self._qiwi_wallet.create_new_balance(currency_alias)

    async def available_balances(self) -> List[AvailableBalance]:
        """
        The request displays account aliases, available for creation in your QIWI Wallet

        """
        return await self._qiwi_wallet.get_available_balances()

    async def set_default_balance(self, currency_alias: str) -> Dict[Any, Any]:
        """
        The request sets up an account for your QIWI Wallet, whose balance will be used for funding
        all payments by default.
        The account must be contained in the list of accounts, you can get the list by calling
        get_list_of_balances method

        :param currency_alias:
        """
        return await self._qiwi_wallet.set_default_balance(currency_alias)

    @deprecated(  # type: ignore
        'QiwiWrapper.to_wallet(...) method is outdated, please use QiwiWrapper.transfer_money(...)'
    )
    async def to_wallet(
        self,
        to_number: str,
        amount: Union[AmountType, str],
        currency: str = '643',
        comment: Optional[str] = None,
    ) -> PaymentInfo:
        return await self.transfer_money(to_phone_number=to_number, amount=amount, comment=comment)

    async def transfer_money(
        self,
        to_phone_number: str,
        amount: Union[AmountType, str],
        comment: Optional[str] = None,
    ) -> PaymentInfo:
        """
        Method for transferring funds to wallet

        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#p2p

        :param to_phone_number: recipient number
        :param amount: the amount of money you want to transfer
        :param comment: payment comment
        """
        return await self._qiwi_wallet.transfer_money(to_phone_number, amount, comment)

    @deprecated(  # type: ignore
        'QiwiWrapper.to_card(...) method is outdated, please use QiwiWrapper.transfer_money_to_card(...)'
    )
    async def to_card(self, trans_sum: Union[float, int], to_card: str) -> PaymentInfo:
        return await self.transfer_money_to_card(card_number=to_card, amount=trans_sum)

    async def transfer_money_to_card(self, card_number: str, *, amount: AmountType) -> PaymentInfo:
        """
        Method for sending funds to the card.

        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#cards
        """
        return await self._qiwi_wallet.transfer_money_to_card(amount, card_number)

    async def predict_commission(self, to_account: str, pay_sum: AmountType) -> Commission:
        """
        Full calc_commission of QIWI Wallet is refunded for payment in favor of the specified provider
        taking into account all tariffs for a given set of payment details.

        :param to_account:
        :param pay_sum:
        :return: Commission object
        """
        return await self._qiwi_wallet.predict_commission(to_account, pay_sum)

    async def get_cross_rates(self) -> List[CrossRate]:
        """
        The method returns the current exchange rates and cross-rates of the QIWI Bank's currencies.

        """
        return await self._qiwi_wallet.get_cross_rates()

    async def payment_by_payment_details(
        self,
        payment_sum: AmountWithCurrency,
        payment_method: PaymentMethod,
        fields: PaymentDetails,
        payment_id: Optional[str] = None,
    ) -> PaymentInfo:
        """
        Payment for services of commercial organizations according to their bank details.

        :param payment_id: payment id, if not transmitted, is used uuid4 by default
        :param payment_sum: a Sum object, which indicates the amount of the payment
        :param payment_method: payment method
        :param fields: payment details
        """
        return await self._qiwi_wallet.payment_by_payment_details(
            payment_sum, payment_method, fields, payment_id
        )

    async def buy_qiwi_master(self) -> PaymentInfo:
        """
        Method for buying QIWI Master package
        To call API methods, you need the QIWI Wallet API token with permissions to do the following:
        1. Management of virtual cards,
        2. Request information about the wallet profile,
        3. View payment history,
        4. Making payments without SMS.
        You can choose these rights when creating a new api token, to use api QIWI Master
        """
        return await self._qiwi_wallet.buy_qiwi_master_package()

    async def issue_qiwi_master_card(self, card_alias: str = 'qvc-cpa') -> Optional[OrderDetails]:
        """
        Issuing a new card using the Qiwi Master API

        When issuing a card, 3, and possibly 3 requests are made, namely,
        according to the following scheme:
            - _create_card_purchase_order - this method creates a request
            - _confirm_qiwi_master_purchase_order - confirms the issue of the card
            - _buy_new_qiwi_master_card - buys a new card,
              if such a card is not free
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#qiwi-master-issue-card
        """
        return await self._qiwi_wallet.issue_qiwi_master_card(card_alias)

    async def register_webhook(self, web_url: str, txn_type: int = 2) -> WebhookInfo:
        """
        This method register a new webhook

        :param web_url: service endpoint
        :param txn_type:  0 => incoming, 1 => outgoing, 2 => all
        :return: Active Hooks
        """
        return await self._qiwi_wallet.register_webhook(web_url, txn_type)

    async def get_current_webhook(self) -> WebhookInfo:
        """
        List of active (active) notification handlers, associated with your wallet can be obtained with this request.
        Since now only one type of hook is used - webhook, then the response contains only one data object
        """
        return await self._qiwi_wallet.get_current_webhook()

    async def send_test_notification(self) -> Dict[str, Any]:
        """
        Use this request to test your webhooks handler.
        Test notification is sent to the address specified during the call register_webhook
        """
        return await self._qiwi_wallet.send_test_webhook_notification()

    async def get_webhook_secret_key(self, hook_id: str) -> str:
        """
        Each notification contains a digital signature of the message, encrypted with a key.
        To obtain a signature verification key, use this request.

        :param hook_id: UUID of webhook
        :return: Base64 encoded key
        """
        return await self._qiwi_wallet.get_webhook_secret_key(hook_id)

    async def delete_current_webhook(self) -> Dict[str, Any]:
        """Method to delete webhook"""
        return await self._qiwi_wallet.delete_current_webhook()

    async def change_webhook_secret(self, hook_id: str) -> str:
        """
        Use this request to change the encryption key for notifications.

        :param hook_id: UUID of webhook
        :return: Base64 encoded key
        """
        return await self._qiwi_wallet.generate_new_webhook_secret(hook_id)

    async def bind_webhook(
        self,
        url: Union[str, 'WebhookURL'],
        transactions_type: int = 2,
        *,
        send_test_notification: bool = False,
        delete_old: bool = False,
    ) -> Tuple[WebhookInfo, str]:
        """
        [NON-API] EXCLUSIVE method to register new webhook or get old

        :param url: service endpoint
        :param transactions_type: 0 => incoming, 1 => outgoing, 2 => all
        :param send_test_notification:  test_qiwi will transfer_money
         you test webhook update
        :param delete_old: boolean, if True - delete old webhook

        :return: Tuple of Hook and Base64-encoded key
        """
        return await self._qiwi_wallet.bind_webhook(
            url,
            transactions_type=transactions_type,
            send_test_notification=send_test_notification,
            delete_old=delete_old,
        )

    async def detect_mobile_operator(self, phone_number: str) -> MobileOperator:
        return await self._qiwi_wallet.detect_mobile_operator(phone_number=phone_number)

    async def reject_bill_by_id(self, bill_id: str) -> Bill:
        """Use this method to cancel unpaid invoice."""
        return await self._p2p_client.reject_p2p_bill(bill_id)

    async def reject_bill(self, bill: Bill) -> Bill:
        return await self._p2p_client.reject_bill(bill)

    async def check_p2p_bill_status(self, bill_id: str) -> str:
        """
        Method for checking the status of a p2p transaction.\n
        Possible transaction types: \n
        WAITING	Bill is waiting for pay	\n
        PAID	Bill was paid	\n
        REJECTED	Bill was rejected\n
        EXPIRED	The bill has expired. Invoice not paid\n
        Docs:
        https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id:
        :return: status of bill
        """
        return await self._p2p_client.get_bill_status(bill_id)

    async def check_if_bill_was_paid(self, bill: Bill) -> bool:
        return await self._p2p_client.check_if_bill_was_paid(bill)

    async def create_p2p_bill(
        self,
        amount: AmountType,
        bill_id: Optional[str] = None,
        comment: Optional[str] = None,
        life_time: Optional[datetime] = None,
        theme_code: Optional[str] = None,
        pay_source_filter: Optional[List[str]] = None,
        customer: Optional[Customer] = None,
    ) -> Bill:
        """
        It is the reliable method for integration.
        Parameters are sent by means of server2server requests with authorization.
        Method allows you to issue an invoice, successful
         response contains payUrl link to redirect client on Payment Form.
        Possible values of pay_source_filter:
          - 'qw'
          - 'card'
          - 'mobile'

        :param amount: amount of payment
        :param bill_id: unique transaction number, if not transmitted,
         generated automatically,
        :param life_time: the date until which the invoice will be available for payment.
        :param comment:
        :param theme_code:
        :param pay_source_filter: When you open the form, the following will be displayed
         only the translation methods specified in this parameter
        :param customer:
        """
        return await self._p2p_client.create_p2p_bill(
            amount, bill_id, comment, life_time, theme_code, pay_source_filter, customer
        )

    async def retrieve_bills(self, rows: int, statuses: str = 'READY_FOR_PAY') -> List[Bill]:
        """
        A method for getting a list of your wallet's outstanding bills.

        The list is built in reverse chronological order.

        By default, the list is paginated with 50 items each,
        but you can specify a different number of elements (no more than 50).

        Filters by billing time can be used in the request,
        the initial account identifier.
        """
        return await self._qiwi_wallet.list_of_invoices(rows, statuses)

    async def pay_the_invoice(self, invoice_uid: str, currency: str) -> InvoiceStatus:
        """
        Execution of unconditional payment of the invoice without SMS-confirmation.

        ! Warning !
        To use this method correctly you need to tick "Проведение платежей без SMS"
        when registering QIWI API and retrieve token

        :param invoice_uid: Bill ID in QIWI system
        :param currency:
        """
        return await self._qiwi_wallet.pay_the_invoice(invoice_uid, currency)

    async def refund_bill(
        self,
        bill_id: Union[str, int],
        refund_id: Union[str, int],
        json_bill_data: Union[PlainAmount, Dict[str, Union[str, int]]],
    ) -> RefundedBill:
        """
        The method allows you to make a refund on a paid invoice.
         in the JSON body of the request for the json_bill_data parameter:
         amount.value - refund amount.
         amount.currency - return currency.
        Can be a dictionary or an OptionalSum object
         Dictionary example: {
        "amount": {
            "currency": "RUB",
            "value": 1
            }
        }

        :param bill_id: unique account identifier in the merchant's system
        :param refund_id: unique identifier of the refund in the merchant's system.
        :param json_bill_data:
        :return: RefundBill object
        """
        return await self._p2p_client.refund_bill(bill_id, refund_id, json_bill_data)

    async def create_p2p_keys(
        self, key_pair_name: str, server_notification_url: Optional[str] = None
    ) -> PairOfP2PKeys:
        """
        Creates a new pair of P2P keys to interact with P2P QIWI API

        :param key_pair_name: P2P token pair name
        :param server_notification_url: endpoint for webhooks
        """
        return await self._p2p_client.create_pair_of_p2p_keys(
            key_pair_name, server_notification_url
        )

    def create_shim_url(self, invoice_uid: str) -> str:
        return self._p2p_client.create_shim_url(invoice_uid)

    async def __aenter__(self) -> 'QiwiWrapper':
        await self._qiwi_wallet.__aenter__()  # type: ignore
        await self._p2p_client.__aenter__()  # type: ignore
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        await self._qiwi_wallet.__aexit__(exc_type, exc_val, exc_tb)
        await self._p2p_client.__aexit__(exc_type, exc_val, exc_tb)

    async def close(self) -> None:
        await self._qiwi_wallet.close()
        await self._p2p_client.close()

    def __new__(
        cls: Type[_T],
        api_access_token: Optional[str] = None,
        phone_number: Optional[str] = None,
        secret_p2p: Optional[str] = None,
        cache_time_in_seconds: Union[float, int] = 0,
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        if not isinstance(api_access_token, str) and not isinstance(secret_p2p, str):  # noqa: W503
            raise RuntimeError('Unable to initialize QiwiWrapper instance without any tokens')

        return super().__new__(cls)



from __future__ import annotations

import abc
import inspect
from copy import deepcopy
from types import TracebackType
from typing import TYPE_CHECKING as MYPY
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple, Type, TypeVar, Union

if MYPY:
    from glQiwiApi.core.request_service import RequestServiceProto  # pragma: no cover

T = TypeVar('T', bound='BaseAPIClient')

RequestServiceFactoryType = Callable[
    ..., Union[Awaitable['RequestServiceProto'], 'RequestServiceProto']
]

_C = TypeVar('_C', bound=Type['APIClientMeta'])


class APIClientMeta(abc.ABCMeta):
    """
    This metaclass fixes issue connected with poor release
    of glQiwiApi 2.0.5 that breaks all down. It's a complicated workaround of building of request service
    "on-flight" and allows users of API client to use custom factories of a request service.

    Due to limitations of aiohttp library that requires to create ClientSession only inside coroutine
    I have to write this metaclass to avoid additional viscous boilerplate code.
    """

    def __new__(
        mcs: Type[_C], name: str, bases: Tuple[Any, ...], attrs: Dict[str, Any], **kwargs: Any
    ) -> _C:
        for key, attribute in attrs.items():
            is_name_mangled = key.startswith('__')
            if is_name_mangled:
                continue

            if not inspect.iscoroutinefunction(attribute):
                continue

            if key in ('close', '_create_request_service', 'create_request_service'):
                continue

            def wrapper(m) -> Any:
                async def check_request_service_before_execute(self, *args: Any, **kw: Any) -> Any:
                    if self._request_service is None:
                        self._request_service = await self.create_request_service()

                    return await m(self, *args, **kw)

                return check_request_service_before_execute

            attrs[key] = wrapper(attrs[key])

        return super().__new__(mcs, name, bases, attrs, **kwargs)


class BaseAPIClient(metaclass=APIClientMeta):
    def __init__(
        self,
        request_service_factory: Optional[RequestServiceFactoryType] = None,
    ):
        self._request_service_factory = request_service_factory
        self._request_service: RequestServiceProto = None  # type: ignore

    async def __aenter__(self):  # type: ignore
        if self._request_service is None:
            self._request_service = await self.create_request_service()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def close(self) -> None:
        if self._request_service is None:
            return None

        await self._request_service.shutdown()

    async def create_request_service(self) -> RequestServiceProto:
        if self._request_service_factory is not None:
            if inspect.iscoroutinefunction(self._request_service_factory):
                return await self._request_service_factory(self)  # type: ignore
            return self._request_service_factory(self)  # type: ignore

        return await self._create_request_service()

    @abc.abstractmethod
    async def _create_request_service(self) -> RequestServiceProto:
        pass

    def __deepcopy__(self: T, memo: Dict[Any, Any]) -> T:
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            try:
                setattr(result, k, deepcopy(v, memo))
            except TypeError:
                pass
        return result



from .invalidation import (
    APIResponsesCacheInvalidationStrategy,
    CacheInvalidationByTimerStrategy,
    CacheInvalidationStrategy,
    UnrealizedCacheInvalidationStrategy,
)
from .storage import InMemoryCacheStorage



from http import HTTPStatus
from typing import Any, ClassVar, Dict, Sequence

from pydantic import Field

from glQiwiApi.core.abc.api_method import Request
from glQiwiApi.qiwi.base import QiwiAPIMethod


class CreateNewBalance(QiwiAPIMethod[Dict[str, Any]]):
    url: ClassVar[str] = 'https://edge.qiwi.com/funding-sources/v2/persons/{phone_number}/accounts'
    http_method: ClassVar[str] = 'POST'
    arbitrary_allowed_response_status_codes: ClassVar[Sequence[int]] = [HTTPStatus.CREATED]

    currency_alias: str = Field(..., alias='alias')

    def build_request(self, **url_format_kw: Any) -> 'Request':
        return Request(
            endpoint=self.url.format(**url_format_kw, **self._get_runtime_path_values()),
            http_method=self.http_method,
            json_payload={'alias': self.currency_alias},
        )



from __future__ import annotations

import asyncio
from concurrent import futures as futures
from concurrent.futures import Future
from typing import Any, Awaitable, Callable, Coroutine, Optional, TypeVar, Union

N = TypeVar("N")

def run_forever_safe(
    loop: asyncio.AbstractEventLoop,
    callback: Optional[Callable[..., Awaitable[N]]] = None,
) -> None: ...
def safe_cancel(
    loop: asyncio.AbstractEventLoop,
    callback: Optional[Callable[..., Awaitable[N]]],
) -> None: ...

AnyExecutor = Union[futures.ThreadPoolExecutor, futures.ProcessPoolExecutor, Optional[None]]

def _cancel_future(
    loop: asyncio.AbstractEventLoop,
    future: asyncio.Future[N],
    executor: AnyExecutor,
) -> None: ...
def _stop_loop(loop: asyncio.AbstractEventLoop) -> None: ...
def take_event_loop(set_debug: bool = False) -> asyncio.AbstractEventLoop: ...
def await_sync(future: Future[N]) -> N: ...
def execute_async_as_sync(
    func: Callable[..., Coroutine[Any, Any, N]], *args: object, **kwargs: object
) -> N: ...

class async_as_sync:  # NOQA
    _async_shutdown_callback: Optional[Callable[..., Awaitable[N]]]
    _sync_shutdown_callback: Optional[Callable[[Any], Any]]
    def __init__(
        self,
        async_shutdown_callback: Optional[Callable[..., Awaitable[N]]] = None,
        sync_shutdown_callback: Optional[Callable[[Any], Any]] = None,
    ) -> None: ...
    def __call__(self, func: Callable[..., Awaitable[N]]) -> Callable[..., N]: ...
    def execute_sync_callback(self, result: Any) -> Any: ...



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Iterator, List, Optional, Union

from pydantic import BaseModel, Field, root_validator

from glQiwiApi.types.base import Base
from glQiwiApi.utils.compat import Literal
from glQiwiApi.yoo_money.exceptions import YooMoneyError, YooMoneyErrorSchema


class Response(Base):
    error: Optional[str] = None

    @root_validator(pre=True)
    def _check_error(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('error') is None:
            return values

        YooMoneyError.raise_most_appropriate_error(YooMoneyErrorSchema(error_code=values['error']))
        return values


class BalanceDetails(Response):
    """object: BalanceDetails"""

    total: float
    available: float
    deposition_pending: Optional[float] = None
    blocked: Optional[float] = None
    debt: Optional[float] = None
    hold: Optional[float] = None


class CardsLinked(BaseModel):
    """object: CardsLinked"""

    pan_fragment: str
    type: str


class AccountInfo(Response):
    """object: AccountInfo"""

    account: str
    """Номер счета"""

    balance: float
    """Баланс счета"""

    currency: str
    """
    Код валюты счета пользователя.
    Всегда 643 (рубль РФ по стандарту ISO 4217).
    """

    identified: bool
    """Верифицирован ли кошелек"""

    account_type: str
    """
    Тип счета пользователя. Возможные значения:
    personal — счет пользователя в ЮMoney;
    professional — профессиональный счет в ЮMoney
    """

    account_status: str
    """
    Статус пользователя. Возможные значения:
    anonymous — анонимный счет;
    named — именной счет;
    identified — идентифицированный счет.
    """

    balance_details: BalanceDetails
    """
    Расширенная информация о балансе.
    По умолчанию этот блок отсутствует.
    Блок появляется, если сейчас или когда-либо ранее были зачисления в очереди
    задолженности, блокировки средств.
    """

    cards_linked: Optional[List[CardsLinked]] = None
    """
    Информация о привязанных банковских картах.
    Если к счету не привязано ни одной карты, параметр отсутствует.
    Если к счету привязана хотя бы одна карта,
    параметр содержит список данных о привязанных картах.
    pan_fragment - string Маскированный номер карты.
    type: string
    Тип карты. Может отсутствовать, если неизвестен. Возможные значения:
    - VISA;
    - MasterCard;
    - AmericanExpress;
    - JCB.
    """


class DigitalGoods(Response):
    """
    Данные о цифровом товаре (пин-коды и бонусы игр, iTunes, Xbox, etc.)
    Поле присутствует при успешном платеже в магазины цифровых товаров.
    Описание формата:
    https://yoomoney.ru/docs/wallet/process-payments/process-payment#digital-goods
    """

    article_id: str = Field(..., alias='merchantArticleId')
    serial: str
    secret: str


class Operation(Response):
    """object: Operation"""

    id: str = Field(..., alias='operation_id')
    """Идентификатор операции."""

    status: str
    """
    Статус платежа (перевода). Может принимать следующие значения:
    - success — платеж завершен успешно;
    - refused — платеж отвергнут получателем или отменен отправителем;
    - in_progress — платеж не завершен,
     перевод не принят получателем или ожидает ввода кода протекции.
    """

    operation_date: datetime = Field(..., alias='datetime')
    """
    Дата и время совершения операции в формате строки
    в ISO формате с часовым поясом UTC.
    """

    title: str
    """
    Краткое описание операции (название магазина или источник пополнения).
    """

    direction: str
    """
    Направление движения средств. Может принимать значения:
    - in (приход);
    - out (расход).
    """

    amount: float
    """Сумма операции."""

    operation_type: str = Field(..., alias='type')
    """Тип операции. Возможные значения:
    payment-shop — исходящий платеж в магазин;
    outgoing-transfer — исходящий P2P-перевод любого типа;
    deposition — зачисление;
    incoming-transfer — входящий перевод или перевод до востребования;
    incoming-transfer-protected — входящий перевод с кодом протекции.
    """

    label: Optional[str] = None
    """Метка платежа.
     Присутствует для входящих и исходящих переводов другим пользователям
     у которых был указан параметр label вызова transfer_money()
     """

    pattern_id: Optional[str] = None
    """
    Идентификатор шаблона, по которому совершен платеж.
    Присутствует только для платежей.
    """

    details: Optional[str] = None


class OperationHistory(Response):
    next_record: Optional[int]
    operations: List[Operation]

    def __iter__(self) -> Iterator[Operation]:  # type: ignore
        for operation in self.operations:
            yield operation


class OperationDetails(Response):
    """object: OperationDetails"""

    @root_validator(pre=True)
    def _extract_amount_and_comment_by_operation_type(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        operation_type: str = values['type']
        if operation_type == 'payment':
            values['amount'] = values['amount_due']
            values['comment'] = values['message']
        return values

    id: str = Field(..., alias='operation_id')
    status: str
    amount: float
    amount_due: Optional[float] = None
    currency: str = Field(..., alias='amount_currency')
    operation_date: datetime = Field(..., alias='datetime')
    operation_type: str = Field(..., alias='type')
    direction: Literal['in', 'out']
    title: str
    details: Optional[str] = None
    digital_goods: Optional[Dict[str, DigitalGoods]] = None
    comment: Optional[str] = None
    label: Optional[str] = None
    answer_datetime: Optional[str] = None
    expires: Optional[datetime] = None
    protection_code: Optional[str] = None
    is_secure: bool = Field(default=False, alias='codepro')
    recipient_type: Optional[str] = None
    recipient: Optional[str] = None
    sender: Optional[str] = None
    fee: Optional[float] = None
    pattern_id: Optional[str] = None


class Wallet(BaseModel):
    """object: Wallet"""

    allowed: bool


class Item(Response):
    """object: Item"""

    item_id: str = Field(..., alias='id')
    """
    Идентификатор привязанной к счету банковской карты.
    Его необходимо указать в методе process-payment для
    совершения платежа выбранной картой.
    """
    pan_fragment: str
    """
    Фрагмент номера банковской карты.
    Поле присутствует только для привязанной банковской карты.
    Может отсутствовать, если неизвестен.
    """
    item_type: str = Field(..., alias='type')
    """
    Тип карты. Может отсутствовать, если неизвестен. Возможные значения:
    Visa;
    MasterCard;
    American Express;
    JCB.
    """


class Card(BaseModel):
    """object: Card"""

    allowed: bool
    csc_required: bool
    items: List[Item]


class MoneySource(BaseModel):
    """
    Список доступных методов для проведения данного платежа.
    Каждый метод содержит набор атрибутов.
    """

    wallet: Optional[Wallet] = None
    """платеж со счета пользователя"""
    cards: Optional[Card] = None
    """платеж с банковских карт, привязанных к счету"""


class RequestPaymentResponse(Response):
    """
    Объект, который вы получаете при вызове make_request_payment.
    При вызове данного метода вы не списываете деньги со своего счёта,
    а условно подготавливаете его к отправке.
    Для отправки денег на счёт используйте метод transfer_money()
    """

    status: str
    request_id: str
    recipient_account_status: str
    fees: Dict[str, float]
    balance: Optional[float] = None
    recipient_account_type: Optional[str] = None
    recipient_identified: bool = False
    recipient_masked_account: Optional[str] = None
    multiple_recipients_found: Optional[str] = None
    contract_amount: Optional[float] = None
    money_source: Optional[MoneySource] = None
    protection_code: Optional[str] = None
    account_unblock_uri: Optional[str] = None
    ext_action_uri: Optional[str] = None


class Payment(Response):
    """object: Payment"""

    status: str
    """
    Код результата выполнения операции. Возможные значения:
    success — успешное выполнение (платеж проведен).
    Это конечное состояние платежа.
    refused — отказ в проведении платежа.
    Причина отказа возвращается в поле error. Это конечное состояние платежа.
    in_progress — авторизация платежа не завершена.
    Приложению следует повторить запрос
    с теми же параметрами спустя некоторое время.
    ext_auth_required — для завершения авторизации платежа
    с использованием банковской карты
    требуется аутентификация по технологии 3‑D Secure.
    все прочие значения — состояние платежа неизвестно.
    Приложению следует повторить запрос с теми же параметрами
    спустя некоторое время.
    """

    payment_id: str
    """
    Идентификатор проведенного платежа.
    Присутствует только при успешном выполнении метода transfer_money().
    """

    credit_amount: Optional[float] = None
    """
    Сумма, поступившая на счет получателя.
    Присутствует при успешном переводе средств
    на счет другого пользователя ЮMoney.
    """

    payer: Optional[str] = None
    """
    Номер счета плательщика.
    Присутствует при успешном переводе средств
    на счет другого пользователя ЮMoney.
    """

    payee: Optional[str] = None
    """
    Номер счета получателя.
    Присутствует при успешном переводе средств
    на счет другого пользователя ЮMoney.
    """

    payee_uid: Union[str, int, None] = None
    """Служебное значение, не фигурирует в документации"""

    invoice_id: Optional[str] = None
    """
    Номер транзакции магазина в ЮMoney.
    Присутствует при успешном выполнении платежа в магазин.
    """

    balance: Optional[float] = None
    """
    Баланс счета пользователя после проведения платежа.
    Присутствует только при выполнении следующих условий:
    - метод выполнен успешно;
    - токен авторизации обладает правом account-info.
    """

    account_unblock_uri: Optional[str] = None
    """
    Адрес, на который необходимо отправить пользователя для разблокировки счета
    Поле присутствует в случае ошибки account_blocked.
    """

    acs_uri: Optional[str] = None

    acs_params: Optional[str] = None

    next_retry: Optional[int] = None
    """
    Рекомендуемое время,
    спустя которое следует повторить запрос, в миллисекундах.
    Поле присутствует при status=in_progress.
    """

    digital_goods: Optional[Dict[str, Dict[str, List[DigitalGoods]]]] = None

    protection_code: Optional[str] = None
    """
    Код протекции, который был сгенерирован,
    если при вызове метода апи transfer_money вы указали protect=True
    при передаче аргументов
    """

    def initialize(self, protection_code: Optional[str]) -> 'Payment':
        self.protection_code = protection_code
        return self


class IncomingTransaction(Response):
    """object: IncomingTransaction"""

    status: str
    protection_code_attempts_available: int
    ext_action_uri: Optional[str] = None



from typing import Any, ClassVar, Dict

from glQiwiApi.core.abc.api_method import APIMethod, ReturningType
from glQiwiApi.core.session.holder import HTTPResponse


class RevokeAPIToken(APIMethod[Dict[Any, Any]]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/revoke'

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> ReturningType:
        return {}



import sys

from glQiwiApi.qiwi.clients.maps.client import QiwiMaps
from glQiwiApi.qiwi.clients.p2p.client import QiwiP2PClient
from glQiwiApi.qiwi.clients.wallet.client import QiwiWallet
from glQiwiApi.qiwi.clients.wrapper import QiwiWrapper

from .core.cache import APIResponsesCacheInvalidationStrategy, InMemoryCacheStorage
from .core.cache.storage import CacheStorage
from .yoo_money import YooMoneyAPI


def default_cache_storage() -> CacheStorage:
    return InMemoryCacheStorage(invalidate_strategy=APIResponsesCacheInvalidationStrategy())


if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata

try:
    __version__ = metadata.version('glQiwiApi')
except metadata.PackageNotFoundError:
    __version__ = '99.99.99'

__all__ = (
    # clients
    'YooMoneyAPI',
    'QiwiMaps',
    'QiwiWallet',
    'QiwiP2PClient',
    'QiwiWrapper',
    # other
    '__version__',
    'default_cache_storage',
)



from __future__ import annotations

import ssl
from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from aiohttp import web

from glQiwiApi.utils.certificates import SSLCertificate

DEFAULT_QIWI_WEBHOOK_PATH = '/webhooks/qiwi/transactions/'
DEFAULT_QIWI_ROUTE_NAME = 'QIWI_TRANSACTIONS'

DEFAULT_QIWI_BILLS_WEBHOOK_PATH = '/webhooks/qiwi/bills/'
DEFAULT_QIWI_BILLS_ROUTE_NAME = 'QIWI_BILLS'


@dataclass()
class ApplicationConfig:
    base_app: Optional[web.Application] = None

    host: str = 'localhost'
    'server host'

    port: int = 8080
    'server port that open for tcp/ip trans.'

    ssl_certificate: Optional[SSLCertificate] = None

    kwargs: Dict[Any, Any] = field(default_factory=dict)

    @property
    def ssl_context(self) -> Optional[ssl.SSLContext]:
        if self.ssl_certificate is None:
            return None
        return self.ssl_certificate.as_ssl_context()


@dataclass()
class RoutesConfig:
    p2p_path: str = DEFAULT_QIWI_BILLS_WEBHOOK_PATH
    standard_qiwi_hook_path: str = DEFAULT_QIWI_WEBHOOK_PATH

    p2p_view_route_name: str = DEFAULT_QIWI_BILLS_ROUTE_NAME
    standard_qiwi_route_name: str = DEFAULT_QIWI_ROUTE_NAME


@dataclass()
class EncryptionConfig:
    secret_p2p_key: str
    base64_encryption_key: Optional[
        str
    ] = None  # taken from QIWI API using QiwiWallet instance by default


@dataclass()
class SecurityConfig:
    check_ip: bool = True


@dataclass()
class HookRegistrationConfig:
    host_or_ip_address: Optional[str] = None


@dataclass
class WebhookConfig:
    encryption: EncryptionConfig
    hook_registration: HookRegistrationConfig = field(default_factory=HookRegistrationConfig)
    app: ApplicationConfig = field(default_factory=ApplicationConfig)
    routes: RoutesConfig = field(default_factory=RoutesConfig)
    security: SecurityConfig = field(default_factory=SecurityConfig)



from typing import ClassVar, cast

from pydantic import Field

from glQiwiApi.core.abc.api_method import ReturningType
from glQiwiApi.core.session.holder import HTTPResponse
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.mobile_operator import MobileOperator
from glQiwiApi.qiwi.exceptions import MobileOperatorCannotBeDeterminedError


class DetectMobileNumber(QiwiAPIMethod[MobileOperator]):
    url: ClassVar[str] = 'https://qiwi.com/mobile/detect.action'
    http_method: ClassVar[str] = 'POST'

    phone_number: str = Field(..., alias='phone')

    @classmethod
    def parse_http_response(cls, response: HTTPResponse) -> ReturningType:
        mobile_operator: MobileOperator = super().parse_http_response(response)
        if mobile_operator.code.value == '2' or mobile_operator.code.name == 'ERROR':
            raise MobileOperatorCannotBeDeterminedError(
                response, custom_message=mobile_operator.message
            )

        return cast(ReturningType, mobile_operator)



from .abc import Pluggable
from .aiogram.polling import AiogramPollingPlugin
from .aiogram.webhook import AiogramWebhookPlugin

__all__ = ('Pluggable', 'AiogramPollingPlugin', 'AiogramWebhookPlugin')



from typing import ClassVar

from glQiwiApi.core.abc.api_method import APIMethod
from glQiwiApi.yoo_money.types import IncomingTransaction


class AcceptIncomingTransfer(APIMethod[IncomingTransaction]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://yoomoney.ru/api/incoming-transfer-accept'

    operation_id: str
    protection_code: str



import re
from typing import Any, Dict, TypeVar, cast

from pydantic import BaseModel

Model = TypeVar('Model', bound=BaseModel)
DEFAULT_EXCLUDE = ('cls', 'self', '__class__')


def filter_dictionary_none_values(dictionary: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Pop NoneType values and convert everything to str, designed?for=params
    :param dictionary: source dict
    :return: filtered dict
    """
    return {k: str(v) for k, v in dictionary.items() if v is not None}


def make_payload(**kwargs: Any) -> Dict[Any, Any]:
    exclude_list = kwargs.pop('exclude', ())
    return {
        key: value
        for key, value in kwargs.items()
        if key not in DEFAULT_EXCLUDE + exclude_list and value is not None
    }


def parse_auth_link(response_data: str) -> str:
    """
    Parse link for getting code, which needs to be entered in the method
    get_access_token
    :param response_data:
    """
    regexp = re.compile(
        r'https://yoomoney.ru/oauth2/authorize[?]requestid[=]\w+'
    )  # pragma: no cover
    return cast(str, re.findall(regexp, str(response_data))[0])  # pragma: no cover



from typing import Any, ClassVar, Dict, Optional, Union

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types import Commission


class PredictCommission(QiwiAPIMethod[Commission]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[str] = 'https://edge.qiwi.com/sinap/providers/{private_card_id}/onlineCommission'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'account': RuntimeValue(),
        'paymentMethod': {'type': 'Account', 'accountId': '643'},
        'purchaseTotals': {'total': {'amount': RuntimeValue(), 'currency': '643'}},
    }

    private_card_id: str = Field(..., path_runtime_value=True)

    invoice_amount: Union[str, int, float] = Field(..., scheme_path='purchaseTotals.total.amount')
    to_account: str = Field(..., scheme_path='account')



import abc
from typing import Any, Dict, List, Optional

from glQiwiApi.core.cache.constants import VALUE_PLACEHOLDER
from glQiwiApi.core.cache.exceptions import CacheExpiredError, CacheValidationError
from glQiwiApi.core.cache.invalidation import (
    CacheInvalidationStrategy,
    UnrealizedCacheInvalidationStrategy,
)
from glQiwiApi.core.cache.utils import embed_cache_time


class CacheStorage(abc.ABC):
    def __init__(self, invalidate_strategy: Optional[CacheInvalidationStrategy] = None):
        if invalidate_strategy is None:
            invalidate_strategy = UnrealizedCacheInvalidationStrategy()
        self._invalidate_strategy: CacheInvalidationStrategy = invalidate_strategy

    @abc.abstractmethod
    async def clear(self) -> None:
        pass

    @abc.abstractmethod
    async def update(self, **kwargs: Any) -> None:
        pass

    @abc.abstractmethod
    async def retrieve(self, key: str) -> Any:
        pass

    @abc.abstractmethod
    async def delete(self, key: str) -> None:
        pass

    @abc.abstractmethod
    async def retrieve_all(self) -> List[Any]:
        ...

    @abc.abstractmethod
    async def contains_similar(self, item: Any) -> bool:
        ...

    def __getitem__(self, item: Any) -> Any:
        return self.retrieve(item)

    def __setitem__(self, key: str, value: Any) -> Any:
        self.update(**{key: value})


class InMemoryCacheStorage(CacheStorage):
    __slots__ = ('_data', '_invalidate_strategy')

    def __init__(self, invalidate_strategy: Optional[CacheInvalidationStrategy] = None):
        CacheStorage.__init__(self, invalidate_strategy)
        self._data: Dict[Any, Any] = {}

    async def clear(self) -> None:
        await self._invalidate_strategy.process_delete()
        self._data.clear()

    async def retrieve_all(self) -> List[Optional[Any]]:
        return [self.retrieve(key) for key in list(self._data.keys())]

    async def update(self, **kwargs: Any) -> None:
        try:
            await self._invalidate_strategy.process_update(**kwargs)
        except CacheValidationError:
            return None
        embedded_data: Dict[Any, Any] = embed_cache_time(**kwargs)
        self._data.update(embedded_data)

    async def retrieve(self, key: str) -> Optional[Any]:
        obj = self._data.get(key)
        if obj is None:
            return obj
        try:
            await self._invalidate_strategy.process_retrieve(obj=obj)
            return obj[VALUE_PLACEHOLDER]
        except CacheExpiredError:
            await self.delete(key)
            return None

    async def delete(self, key: str) -> None:
        del self._data[key]

    async def contains_similar(self, item: Any) -> bool:
        return await self._invalidate_strategy.check_is_contains_similar(self, item)

    def __del__(self) -> None:
        del self._data



from __future__ import annotations

import datetime
from typing import List, Optional

from pydantic import Field

from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.base import Base, HashableBase


class QiwiMasterCardRequisites(Base):
    status: str
    cvv: str
    pan: str


class OrderDetails(Base):
    order_id: str = Field(..., alias='id')
    card_alias: str = Field(..., alias='cardAlias')
    status: str
    price: Optional[AmountWithCurrency] = None
    card_id: Optional[str] = Field(alias='cardId', default=None)


class CardCredentials(Base):
    qvx_id: int = Field(..., alias='id')
    masked_pan: str = Field(..., alias='maskedPan')
    status: str
    card_expire: datetime.datetime = Field(..., alias='cardExpire')
    card_type: str = Field(..., alias='cardType')
    card_alias: Optional[str] = Field(..., alias='cardAlias')
    activated: datetime.datetime
    sms_recender: Optional[datetime.datetime] = Field(..., alias='smsResended')
    post_number: Optional[str] = Field(default=None, alias='postNumber')
    blocked_date: Optional[datetime.datetime] = Field(default=None, alias='blockedDate')
    card_id: int = Field(..., alias='cardId')
    txn_id: int = Field(..., alias='txnId')
    card_expire_month: str = Field(..., alias='cardExpireMonth')
    card_expire_year: str = Field(..., alias='cardExpireYear')


class Requisite(Base):
    name: str
    value: str


class Details(Base):
    info: str
    description: str
    tariff_link: str = Field(..., alias='tariffLink')
    offer_link: str = Field(..., alias='offerLink')
    features: List[str]
    requisites: List[Requisite]


class CardInfo(Base):
    id_: int = Field(..., alias='id')
    name: str
    alias: str
    price: AmountWithCurrency
    period: str
    type_: str = Field(..., alias='type')
    details: Details


class Card(HashableBase):
    details: CardCredentials = Field(..., alias='qvx')
    balance: Optional[AmountWithCurrency] = None
    info: CardInfo



from typing import Any, ClassVar, Dict, Optional

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.p2p.types import PairOfP2PKeys


class CreateP2PKeyPair(QiwiAPIMethod[PairOfP2PKeys]):
    http_method: ClassVar[str] = 'POST'
    url: ClassVar[
        str
    ] = 'https://api.qiwi.com/partner/bill/v1/bills/widgets-api/api/p2p/protected/keys/create'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'keysPairName': RuntimeValue(),
        'serverNotificationsUrl': RuntimeValue(),
    }

    key_pair_name: str = Field(..., scheme_path='keysPairName')
    server_notification_url: Optional[str] = Field(..., scheme_path='serverNotificationsUrl')



from __future__ import annotations

import typing as t

from aiohttp import web
from aiohttp.web import _run_app

from glQiwiApi.core.event_fetching.webhooks.config import ApplicationConfig
from glQiwiApi.plugins import Pluggable

if t.TYPE_CHECKING:
    from glQiwiApi.utils.compat import Dispatcher

ListOfRoutes = t.List[web.ResourceRoute]
SubApps = t.List[t.Tuple[str, web.Application, ListOfRoutes]]

DEFAULT_TELEGRAM_WEBHOOK_PATH_PREFIX = '/tg/webhooks'
DEFAULT_TELEGRAM_WEBHOOK_PATH = '/bot/{token}'
DEFAULT_TELEGRAM_WEBHOOK_ROUTE_NAME = 'webhook_handler'

# This one was created only for testing purposes
run_app = _run_app


class SSLCertificateIsMissingError(Exception):
    pass


class AiogramWebhookPlugin(Pluggable):
    """
    Managing loading webhooks of aiogram together with QiwiWrapper

    """

    def __init__(
        self,
        dispatcher: Dispatcher,
        host: str,
        path: str = DEFAULT_TELEGRAM_WEBHOOK_PATH,
        prefix: str = DEFAULT_TELEGRAM_WEBHOOK_PATH_PREFIX,
        route_name: str = DEFAULT_TELEGRAM_WEBHOOK_ROUTE_NAME,
        app_config: ApplicationConfig = ApplicationConfig(),
        **kwargs: t.Any,
    ) -> None:
        """

        :param dispatcher: instance of aiogram class Dispatcher
        :param route_name: name of aiogram webhook router
        :param sub_apps: list of tuples(prefix as string, web.Application)
        """
        from aiogram.dispatcher.webhook import WebhookRequestHandler

        self._path = prefix + path.format(token=dispatcher.bot._token)
        self._host = host
        self._dispatcher = dispatcher
        self._prefix = prefix

        self._app = web.Application()
        self._app_config = app_config
        self._app.router.add_route('*', self._path, WebhookRequestHandler, name=route_name)
        self._app['BOT_DISPATCHER'] = self._dispatcher
        self._set_webhook_kwargs = kwargs

        if self._app_config.ssl_certificate is None:
            raise SSLCertificateIsMissingError(
                "Webhooks won't work without ssl_certificate. "
                'To fix it, please transmit ssl_certificate to ApplicationConfig to TelegramWebhookPlugin'
            )

    async def install(self, ctx: t.Dict[t.Any, t.Any]) -> None:
        """
        A method that connects the main application, and the proxy in the form of a telegram

        :param ctx: keyword arguments, which contains application and host
        """
        await self._set_telegram_webhook(ctx)
        await run_app(
            self._app,
            host=self._app_config.host,
            port=self._app_config.port,
            ssl_context=self._app_config.ssl_certificate.as_ssl_context(),  # type: ignore
        )

    async def _set_telegram_webhook(self, ctx: t.Dict[t.Any, t.Any]) -> None:
        """
        You can override this method to correctly setup webhooks with aiogram
        self.dispatcher.bot.set_webhook(...)

        """
        url = self._host + self._path
        await self._dispatcher.bot.set_webhook(
            url=url,
            certificate=self._app_config.ssl_certificate.as_input_file(),  # type: ignore
            **self._set_webhook_kwargs,
        )

    async def shutdown(self) -> None:
        await self._dispatcher.storage.close()
        await self._dispatcher.storage.wait_closed()
        await self._dispatcher.bot.session.close()



from __future__ import annotations

import abc
import asyncio
import inspect
import logging
import operator
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

from glQiwiApi.qiwi.clients.p2p.types import BillWebhook

from ...qiwi.clients.wallet.types.transaction import Transaction
from ...qiwi.clients.wallet.types.webhooks import TransactionWebhook
from .. import Handler
from .filters import BaseFilter, LambdaBasedFilter

logger = logging.getLogger('glQiwiApi.dispatcher')


class SkipHandler(Exception):
    pass


class CancelHandler(Exception):
    pass


Event = TypeVar('Event')
F = TypeVar('F', bound=Callable[..., Any])


class BaseDispatcher(abc.ABC):
    def __init__(self) -> None:
        self.exception_handler = HandlerCollection(Exception)

    async def process_event(self, event: Event, *args: Any) -> None:
        """
        Feed handlers with event.

        :param event: any object that will be propagated to handlers
        """
        coroutines = [h.notify(event, *args) for h in self.__all_handlers__]
        results = await asyncio.gather(*coroutines, return_exceptions=True)
        for result in results:
            if self.exception_handler.is_empty:
                break
            if not isinstance(result, Exception):
                continue

            await self.exception_handler.notify(result, *args)

    @property
    @abc.abstractmethod
    def __all_handlers__(self) -> Sequence[HandlerCollection[Any]]:
        """Return all registered handlers except error handlers"""


class QiwiDispatcher(BaseDispatcher):
    def __init__(self) -> None:
        super().__init__()
        self.transaction_handler = HandlerCollection(Transaction, TransactionWebhook)
        self.bill_handler = HandlerCollection(BillWebhook)

    @property
    def __all_handlers__(self) -> Sequence[HandlerCollection[Any]]:
        return self.bill_handler, self.transaction_handler


class HandlerCollection(Generic[Event]):
    def __init__(self, *event_types: Type[Event], once: bool = True) -> None:
        self._handlers: List[EventHandler[Event]] = []
        self._once = once
        self._event_filter: Callable[[Event], bool] = lambda e: isinstance(e, event_types)

    async def notify(self, event: Event, *args: Any) -> None:
        if not self._event_filter(event):
            return None
        for handler in self._handlers:
            try:
                await handler.check_then_execute(event, *args)
                if self._once:
                    break
            except SkipHandler:
                continue
            except CancelHandler:
                break

    def __call__(self, *filters: BaseFilter[Event]) -> Callable[[F], F]:
        def decorator(callback: F) -> F:
            self.register_handler(callback, *filters)
            return callback

        return decorator

    def register_handler(
        self,
        event_handler: Union[Callable[..., Awaitable[Any]], Type[Handler[Event]]],
        *filters: Union[Callable[[Event], bool], BaseFilter[Event]],
    ) -> None:
        """
        Add new event handler.
        (!) Allows chain addition.

        :param event_handler: event handler, low order function
         which works with events
        :param filters: filter for low order function execution
        """
        generated_filters: List[BaseFilter[Event]] = []
        for filter_ in filters:
            if inspect.isfunction(filter_):
                generated_filters.append(LambdaBasedFilter(filter_))
            else:
                generated_filters.append(cast(BaseFilter[Event], filter_))

        self._handlers.append(EventHandler(event_handler, *generated_filters))

    def remove_handler(self, handler: EventHandler[Event]) -> None:
        self._handlers.remove(handler)

    def __iter__(self) -> Iterable[EventHandler[Event]]:
        return iter(self._handlers)

    @property
    def is_empty(self) -> bool:
        return len(self._handlers) == 0


class EventHandler(Generic[Event]):
    """
    Event handler, which executing and working with handlers

    """

    def __init__(
        self,
        handler: Union[Callable[..., Awaitable[Any]], Type[Handler[Event]]],
        *filters: BaseFilter[Event],
    ) -> None:
        self._handler = handler
        self._filters = tuple(filter(lambda f: operator.not_(operator.eq(f, None)), filters))

    async def check_then_execute(self, event: Event, *args: Any) -> Optional[Any]:
        """Check event, apply all filters and then pass on to handler"""
        for f in self._filters:
            if not await f.check(event):
                break
        else:
            return await self._handler(event, *args)
        return None  # hint for mypy



import time
from typing import Any, ClassVar, Dict

from pydantic import Field

from glQiwiApi.core.abc.api_method import RuntimeValue
from glQiwiApi.qiwi.base import QiwiAPIMethod
from glQiwiApi.qiwi.clients.wallet.types.qiwi_master import OrderDetails


class BuyQiwiMasterCard(QiwiAPIMethod[OrderDetails]):
    url: ClassVar[str] = 'https://edge.qiwi.com/sinap/api/v2/terms/32064/payments'
    http_method: ClassVar[str] = 'POST'

    json_payload_schema: ClassVar[Dict[str, Any]] = {
        'id': RuntimeValue(default_factory=lambda: str(int(time.time() * 1000))),
        'sum': {'amount': RuntimeValue(default=99), 'currency': '643'},
        'paymentMethod': {'type': 'Account', 'accountId': '643'},
        'fields': {'account': RuntimeValue(), 'order_id': RuntimeValue()},
    }

    phone_number: str = Field(..., schema_path='fields.account')
    order_id: str = Field(..., schema_path='fields.order_id')



from __future__ import annotations

from contextlib import suppress
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from glQiwiApi.core.abc.base_api_client import BaseAPIClient, RequestServiceFactoryType
from glQiwiApi.core.request_service import RequestService, RequestServiceProto
from glQiwiApi.core.session import AiohttpSessionHolder
from glQiwiApi.ext.webhook_url import WebhookURL
from glQiwiApi.qiwi.clients.wallet.methods.authenticate_wallet import AuthenticateWallet
from glQiwiApi.qiwi.clients.wallet.methods.check_restriction import GetRestrictions
from glQiwiApi.qiwi.clients.wallet.methods.create_new_balance import CreateNewBalance
from glQiwiApi.qiwi.clients.wallet.methods.detect_mobile_number import DetectMobileNumber
from glQiwiApi.qiwi.clients.wallet.methods.fetch_statistics import FetchStatistics
from glQiwiApi.qiwi.clients.wallet.methods.get_account_info import GetUserProfile
from glQiwiApi.qiwi.clients.wallet.methods.get_available_balances import GetAvailableBalances
from glQiwiApi.qiwi.clients.wallet.methods.get_balances import GetBalances
from glQiwiApi.qiwi.clients.wallet.methods.get_cards import GetBoundedCards
from glQiwiApi.qiwi.clients.wallet.methods.get_cross_rates import GetCrossRates
from glQiwiApi.qiwi.clients.wallet.methods.get_identification import GetIdentification
from glQiwiApi.qiwi.clients.wallet.methods.get_limits import ALL_LIMIT_TYPES, GetLimits
from glQiwiApi.qiwi.clients.wallet.methods.get_receipt import GetReceipt
from glQiwiApi.qiwi.clients.wallet.methods.history import MAX_HISTORY_LIMIT, GetHistory
from glQiwiApi.qiwi.clients.wallet.methods.list_of_invoices import GetListOfInvoices
from glQiwiApi.qiwi.clients.wallet.methods.pay_invoice import PayInvoice
from glQiwiApi.qiwi.clients.wallet.methods.payment_by_details import MakePaymentByDetails
from glQiwiApi.qiwi.clients.wallet.methods.predict_comission import PredictCommission
from glQiwiApi.qiwi.clients.wallet.methods.qiwi_master.buy_qiwi_card import BuyQiwiMasterCard
from glQiwiApi.qiwi.clients.wallet.methods.qiwi_master.buy_qiwi_master import BuyQIWIMasterPackage
from glQiwiApi.qiwi.clients.wallet.methods.qiwi_master.confirm_qiwi_master import (
    ConfirmQiwiMasterPurchaseOrder,
)
from glQiwiApi.qiwi.clients.wallet.methods.qiwi_master.create_card_purchase_order import (
    CreateCardPurchaseOrder,
)
from glQiwiApi.qiwi.clients.wallet.methods.reveal_card_id import RevealCardID
from glQiwiApi.qiwi.clients.wallet.methods.set_default_balance import SetDefaultBalance
from glQiwiApi.qiwi.clients.wallet.methods.transaction_info import GetTransactionInfo
from glQiwiApi.qiwi.clients.wallet.methods.transfer_money import TransferMoney
from glQiwiApi.qiwi.clients.wallet.methods.transfer_money_to_card import TransferMoneyToCard
from glQiwiApi.qiwi.clients.wallet.methods.webhook.change_webhook_secret import (
    GenerateWebhookSecret,
)
from glQiwiApi.qiwi.clients.wallet.methods.webhook.delete_current_webhook import DeleteWebhook
from glQiwiApi.qiwi.clients.wallet.methods.webhook.get_current_webhook import GetCurrentWebhook
from glQiwiApi.qiwi.clients.wallet.methods.webhook.get_webhook_secret import GetWebhookSecret
from glQiwiApi.qiwi.clients.wallet.methods.webhook.register_webhook import RegisterWebhook
from glQiwiApi.qiwi.clients.wallet.methods.webhook.send_test_notification import (
    SendTestWebhookNotification,
)
from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.arbitrary import File
from glQiwiApi.utils.validators import PhoneNumber, String

from ...exceptions import QiwiAPIError
from ..p2p.types import Bill, InvoiceStatus
from .methods.get_nickname import GetNickName
from .methods.qiwi_master.block_card import BlockQiwiMasterCard
from .methods.qiwi_master.get_card_requisites import GetQiwiMasterCardRequisites
from .methods.qiwi_master.get_statement import GetQiwiMasterStatement
from .methods.qiwi_master.rename_card import RenameQiwiMasterCard
from .methods.qiwi_master.unblock_card import UnblockQiwiMasterCard
from .types import (
    Balance,
    Card,
    Commission,
    CrossRate,
    History,
    Identification,
    Limit,
    OrderDetails,
    PaymentDetails,
    PaymentInfo,
    PaymentMethod,
    Restriction,
    Source,
    Statistic,
    Transaction,
    TransactionType,
    UserProfile,
    WebhookInfo,
)
from .types.balance import AvailableBalance
from .types.mobile_operator import MobileOperator
from .types.nickname import NickName
from .types.qiwi_master import QiwiMasterCardRequisites

AmountType = Union[int, float]


class QiwiWallet(BaseAPIClient):
    _phone_number = PhoneNumber(maxsize=15, minsize=11, optional=True)
    _api_access_token = String(optional=False)

    def __init__(
        self,
        api_access_token: str,
        phone_number: Optional[str] = None,
        request_service_factory: Optional[RequestServiceFactoryType] = None,
    ) -> None:
        """
        :param api_access_token: QIWI API token received from https://qiwi.com/api
        :param phone_number: your phone number starting with +
        """
        self._api_access_token = api_access_token
        self._phone_number = phone_number

        BaseAPIClient.__init__(self, request_service_factory)

    async def _create_request_service(self) -> RequestServiceProto:
        from glQiwiApi import __version__

        return RequestService(
            session_holder=AiohttpSessionHolder(
                headers={
                    'Content-Type': 'application/json',
                    'Accept': 'application/json',
                    'Authorization': f'Bearer {self._api_access_token}',
                    'Host': 'edge.qiwi.com',
                    'User-Agent': f'glQiwiApi/{__version__}',
                }
            )
        )

    async def get_profile(
        self,
        include_auth_info: bool = True,
        include_contract_info: bool = True,
        include_user_info: bool = True,
    ) -> UserProfile:
        return await self._request_service.execute_api_method(
            GetUserProfile(
                include_auth_info=include_auth_info,
                include_contract_info=include_contract_info,
                include_user_info=include_user_info,
            )
        )

    async def authenticate(
        self,
        birth_date: str,
        first_name: str,
        last_name: str,
        middle_name: str,
        passport: str,
        oms: Optional[str] = None,
        inn: Optional[str] = None,
        snils: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """
        This request allows you to transfer data to identify your QIWI wallet.
        It is allowed to identify no more than 5 wallets per owner

        To identify the wallet, you must transfer your full name, passport series number and date of birth.
        If the data has been verified, then the response will display
        your TIN and simplified wallet identification will be installed.
        If the data has not been verified,
        the wallet remains in the "Minimum" status.

        :param birth_date: Date of birth as a format string 1998-02-11
        :param first_name: First name
        :param last_name: Last name
        :param middle_name: Middle name
        :param passport: Series / Number of the passport. Ex: 4400111222
        :param oms:
        :param snils:
        :param inn:
        """
        return await self._request_service.execute_api_method(
            AuthenticateWallet(
                birth_date=birth_date,
                first_name=first_name,
                last_name=last_name,
                middle_name=middle_name,
                passport=passport,
                oms=oms,
                inn=inn,
                snils=snils,
            )
        )

    async def get_identification(self) -> Identification:
        """
        This method allows get your wallet identification data
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#ident
        """
        return await self._request_service.execute_api_method(
            GetIdentification(),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def get_limits(self, limit_types: Sequence[str] = ALL_LIMIT_TYPES) -> Dict[str, Limit]:
        """
        Function for getting limits on the qiwi wallet account
        Returns wallet limits as a list,
        if there is no limit for a certain country, then it does not include it in the list
        payload of limit types must be dict in format like array of strings {
            "types[0]": "Some type",
            "types[1]": "some other type",
            "types[n]": "n type"
        }

        Detailed documentation:

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#limits
        """
        return await self._request_service.execute_api_method(
            GetLimits(limit_types=limit_types), phone_number=self.phone_number_without_plus_sign
        )

    async def get_restrictions(self) -> List[Restriction]:
        """
        Method to check limits on your qiwi wallet
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#restrictions

        :return: List where the dictionary is located with restrictions,
         if there are no restrictions, it returns an empty list
        """
        return await self._request_service.execute_api_method(
            GetRestrictions(),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def history(
        self,
        rows: int = MAX_HISTORY_LIMIT,
        transaction_type: TransactionType = TransactionType.ALL,
        sources: Optional[List[Source]] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        next_txn_date: Optional[datetime] = None,
        next_txn_id: Optional[int] = None,
    ) -> History:
        """
        Method for receiving transactions history on the account
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param rows: number of operation_history you want to receive
        :param transaction_type: The type of operations in the report for selection.
        :param sources: List of payment sources, for filter
        :param start_date: The starting date for searching for payments. Used only in conjunction with end_date.
        :param end_date: the end date of the search for payments. Used only in conjunction with start_date.
        :param next_txn_id:
        :param next_txn_date:
        """
        return await self._request_service.execute_api_method(
            GetHistory(
                rows=rows,
                transaction_type=transaction_type,
                sources=sources,
                start_date=start_date,
                end_date=end_date,
                next_txn_id=next_txn_id,
                next_txn_date=next_txn_date,
            ),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def check_whether_transaction_exists(
        self, check_fn: Callable[[Transaction], bool], rows_num: int = MAX_HISTORY_LIMIT
    ) -> bool:
        """
        [ NON API METHOD ]

        Method for verifying a transaction.
        This method uses self.history (rows = rows) "under the hood" to check payment.

        For a little optimization, you can decrease rows by setting it,
        however, this does not guarantee the correct result

        Correct values for the transaction_type parameter:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'

        """
        history = await self.history(rows=rows_num)
        for transaction in history:
            if check_fn(transaction):
                return True
        return False

    async def fetch_statistics(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        operation: TransactionType = TransactionType.ALL,
        sources: Optional[List[str]] = None,
    ) -> Statistic:
        """
        This query is used to get summary statistics
        by the amount of payments for a given period.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param start_date:The start date of the statistics period.
        :param end_date: End date of the statistics period.
        :param operation: The type of operations taken into account when calculating statistics.
         Allowed values:
            ALL - все операции,
            IN - только пополнения,
            OUT - только платежи,
            QIWI_CARD - только платежи по картам QIWI (QVC, QVP).
            По умолчанию ALL.
        :param sources: The sources of payments
            QW_RUB - рублевый счет кошелька,
            QW_USD - счет кошелька в долларах,
            QW_EUR - счет кошелька в евро,
            CARD - привязанные и непривязанные к кошельку банковские карты,
            MK - счет мобильного оператора. Если не указан,
            учитываются все источники платежа.
        """
        kw = {
            'start_date': start_date,
            'end_date': end_date,
            'operation': operation,
            'sources': sources,
        }

        return await self._request_service.execute_api_method(
            FetchStatistics(**{k: v for k, v in kw.items() if v is not None}),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def get_transaction_info(
        self, transaction_id: Union[str, int], transaction_type: TransactionType
    ) -> Transaction:
        """
        Method for obtaining complete information about a transaction

        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#txn_info

        :param transaction_id:
        :param transaction_type: only IN or OUT
        :return: Transaction object
        """
        return await self._request_service.execute_api_method(
            GetTransactionInfo(transaction_id=transaction_id, transaction_type=transaction_type)
        )

    async def get_receipt(
        self,
        transaction_id: Union[str, int],
        transaction_type: TransactionType,
        file_format: str = 'PDF',
    ) -> File:
        """
        Method for receiving a receipt in byte format or file. \n
        Possible transaction_type values:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'

        :param transaction_id: transaction id, can be obtained by calling the transfer_money method,
         transfer_money_to_card
        :param transaction_type: type of transaction: 'IN', 'OUT', 'QIWI_CARD'
        :param file_format: format of file(JPEG or PDF)
        """
        return await self._request_service.execute_api_method(
            GetReceipt(
                transaction_id=transaction_id,
                transaction_type=transaction_type,
                file_format=file_format,
            )
        )

    async def get_list_of_balances(self) -> List[Balance]:
        """
        The request gets the current account balances of your QIWI Wallet.
        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#balances_list

        """
        return await self._request_service.execute_api_method(
            GetBalances(), phone_number=self.phone_number_without_plus_sign
        )

    async def get_balance(self, *, account_number: int = 1) -> AmountWithCurrency:
        resp: List[Balance] = await self._request_service.execute_api_method(
            GetBalances(),
            phone_number=self.phone_number_without_plus_sign,
        )
        return resp[account_number - 1].balance  # type: ignore

    async def create_new_balance(self, currency_alias: str) -> Dict[str, Any]:
        """
        The request creates a new account and balance in your QIWI Wallet

        :param currency_alias: New account alias
        """
        return await self._request_service.execute_api_method(
            CreateNewBalance(currency_alias=currency_alias),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def detect_mobile_operator(self, phone_number: str) -> MobileOperator:
        """https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#search-providers"""
        return await self._request_service.execute_api_method(
            DetectMobileNumber(phone_number=phone_number)
        )

    async def get_list_of_cards(self, card_alias: Optional[str] = None) -> List[Card]:
        return await self._request_service.execute_api_method(
            GetBoundedCards(card_alias=card_alias)
        )

    async def get_available_balances(self) -> List[AvailableBalance]:
        """
        The request displays account aliases, available for creation in your QIWI Wallet

        """
        return await self._request_service.execute_api_method(
            GetAvailableBalances(), phone_number=self.phone_number_without_plus_sign
        )

    async def set_default_balance(self, account_alias: str) -> Dict[Any, Any]:
        """
        The request sets up an account for your QIWI Wallet, whose balance will be used for funding
        all payments by default.
        The account must be contained in the list of accounts, you can get the list by calling
        get_list_of_balances method

        :param account_alias:
        """
        return await self._request_service.execute_api_method(
            SetDefaultBalance(account_alias=account_alias),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def transfer_money(
        self,
        to_phone_number: str,
        amount: Union[AmountType, str],
        comment: Optional[str] = None,
    ) -> PaymentInfo:
        """
        Method for transferring funds to wallet

        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#p2p

        :param to_phone_number: recipient number
        :param amount: the amount of money you want to transfer
        :param comment: payment comment
        """
        return await self._request_service.execute_api_method(
            TransferMoney(amount=amount, comment=comment, to_wallet=to_phone_number)
        )

    async def transfer_money_to_card(
        self, amount: AmountType, card_number: str, **kwargs: Any
    ) -> PaymentInfo:
        """
        Method for sending funds to the card.

        More detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#cards
        """
        private_card_id = await self.detect_card_id(card_number=card_number)
        return await self._request_service.execute_api_method(
            TransferMoneyToCard(
                private_card_id=private_card_id,
                amount=amount,
                card_number=card_number,
                kwargs=kwargs,
            )
        )

    async def detect_card_id(self, card_number: str) -> str:
        """https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards"""
        return await self._request_service.execute_api_method(
            RevealCardID(card_number=card_number)
        )

    async def predict_commission(self, to_account: str, invoice_amount: AmountType) -> Commission:
        """
        Full calc_commission of QIWI Wallet is refunded for payment in favor of the specified provider
        taking into account all tariffs for a given set of payment details.

        :param to_account:
        :param invoice_amount:
        :return: Commission object
        """
        card_code = '99' if len(to_account) <= 15 else None
        if card_code is None:
            card_code = await self.detect_card_id(to_account)

        return await self._request_service.execute_api_method(
            PredictCommission(
                private_card_id=card_code, invoice_amount=invoice_amount, to_account=to_account
            )
        )

    async def get_cross_rates(self) -> List[CrossRate]:
        """
        The method returns the current exchange rates and cross-rates of the QIWI Bank's currencies.

        """
        return await self._request_service.execute_api_method(GetCrossRates())

    async def payment_by_payment_details(
        self,
        payment_sum: AmountWithCurrency,
        payment_method: PaymentMethod,
        fields: PaymentDetails,
        payment_id: Optional[str] = None,
    ) -> PaymentInfo:  # pragma: no cover
        """
        Payment for services of commercial organizations according to their bank details.

        :param payment_id: payment id, if not transmitted, is used uuid4 by default
        :param payment_sum: a Sum object, which indicates the amount of the payment
        :param payment_method: payment method
        :param fields: payment details
        """
        return await self._request_service.execute_api_method(
            MakePaymentByDetails(
                payment_sum=payment_sum,
                payment_method=payment_method,
                details=fields,
                payment_id=payment_id,
            )
        )

    async def buy_qiwi_master_package(self) -> PaymentInfo:  # pragma: no cover
        """
        Method for buying QIWI Master package
        To call API methods, you need the QIWI Wallet API token with permissions to do the following:
        1. Management of virtual cards,
        2. Request information about the wallet profile,
        3. View payment history,
        4. Making payments without SMS.
        You can choose these rights when creating a new api token, to use api QIWI Master
        """
        return await self._request_service.execute_api_method(
            BuyQIWIMasterPackage(phone_number=self.phone_number_without_plus_sign)
        )

    async def issue_qiwi_master_card(
        self, card_alias: str = 'qvc-cpa', over_the_limit: bool = False
    ) -> OrderDetails:  # pragma: no cover
        """
        Issuing a new card using the Qiwi Master API

        When issuing a card, 3, and possibly 3 requests are made, namely,
        according to the following scheme:
            - _create_card_purchase_order - this method creates a request
            - _confirm_qiwi_master_purchase_order - confirms the issue of the card
            - _buy_new_qiwi_master_card - buys a new card,
              if such a card is not free
        Detailed documentation:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#qiwi-master-issue-card
        """
        order_details = await self._confirm_qiwi_master_purchase_order(card_alias)
        if not over_the_limit or order_details.status == 'COMPLETED':
            return order_details
        return await self._buy_new_qiwi_master_card(order_id=order_details.order_id)

    async def _confirm_qiwi_master_purchase_order(
        self, card_alias: str = 'qvc-cpa'
    ) -> OrderDetails:  # pragma: no cover
        details = await self._create_card_purchase_order(card_alias)
        return await self._request_service.execute_api_method(
            ConfirmQiwiMasterPurchaseOrder(order_id=details.order_id)
        )

    async def _create_card_purchase_order(
        self, card_alias: str = 'qvc-cpa'
    ) -> OrderDetails:  # pragma: no cover
        return await self._request_service.execute_api_method(
            CreateCardPurchaseOrder(card_alias=card_alias),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def _buy_new_qiwi_master_card(self, order_id: str) -> OrderDetails:  # pragma: no cover
        return await self._request_service.execute_api_method(
            BuyQiwiMasterCard(order_id=order_id, phone_number=self.phone_number_without_plus_sign)
        )

    async def get_qiwi_master_card_statement(
        self, card_id: str, from_date: datetime, till_date: datetime
    ) -> File:  # pragma: no cover
        return await self._request_service.execute_api_method(
            GetQiwiMasterStatement(card_id=card_id, from_date=from_date, till_date=till_date),
            phone_number=self.phone_number_without_plus_sign,
        )

    async def block_qiwi_master_card(self, card_id: str) -> None:  # pragma: no cover
        await self._request_service.execute_api_method(
            BlockQiwiMasterCard(card_id=card_id, phone_number=self.phone_number_without_plus_sign)
        )

    async def unblock_qiwi_master_card(self, card_id: str) -> Dict[str, Any]:  # pragma: no cover
        return await self._request_service.execute_api_method(
            UnblockQiwiMasterCard(
                card_id=card_id, phone_number=self.phone_number_without_plus_sign
            )
        )

    async def get_qiwi_master_card_requisites(
        self, card_id: str, operation_id: Optional[str] = None
    ) -> QiwiMasterCardRequisites:  # pragma: no cover
        return await self._request_service.execute_api_method(
            GetQiwiMasterCardRequisites(card_id=card_id, operation_id=operation_id)
        )

    async def rename_qiwi_master_card(
        self, card_id: str, alias: str
    ) -> Dict[str, Any]:  # pragma: no cover
        return await self._request_service.execute_api_method(
            RenameQiwiMasterCard(card_id=card_id, alias=alias)
        )

    async def get_nickname(self) -> NickName:
        return await self._request_service.execute_api_method(
            GetNickName(), phone_number=self.phone_number_without_plus_sign
        )

    async def list_of_invoices(self, rows: int, statuses: str = 'READY_FOR_PAY') -> List[Bill]:
        """
        A method for getting a list of your wallet's outstanding bills.

        The list is built in reverse chronological order.

        By default, the list is paginated with 50 items each,
        but you can specify a different number of elements (no more than 50).

        Filters by billing time can be used in the request,
        the initial account identifier.
        """
        return await self._request_service.execute_api_method(
            GetListOfInvoices(rows=rows, statuses=statuses)
        )

    async def pay_the_invoice(self, invoice_uid: str, currency: str) -> InvoiceStatus:
        """
        Execution of unconditional payment of the invoice without SMS-confirmation.

        ! Warning !
        To use this method correctly you need to tick "Проведение платежей без SMS"
        when registering QIWI API and retrieve token

        :param invoice_uid: Bill ID in QIWI system
        :param currency:
        """
        return await self._request_service.execute_api_method(
            PayInvoice(invoice_uid=invoice_uid, currency=currency)
        )

    async def register_webhook(self, url: str, txn_type: int = 2) -> WebhookInfo:
        """
        This method register a new webhook

        :param url: service endpoint
        :param txn_type:  0 => incoming, 1 => outgoing, 2 => all
        :return: Active Hooks
        """
        return await self._request_service.execute_api_method(
            RegisterWebhook(webhook_url=url, txn_type=txn_type)
        )

    async def get_current_webhook(self) -> WebhookInfo:
        """
        List of active (active) notification handlers, associated with your wallet can be obtained with this request.
        Since now only one type of hook is used - webhook, then the response contains only one data object
        """
        return await self._request_service.execute_api_method(GetCurrentWebhook())

    async def send_test_webhook_notification(self) -> Dict[Any, Any]:
        """
        Use this request to test your webhooks handler.
        Test notification is sent to the address specified during the call register_webhook
        """
        return await self._request_service.execute_api_method(SendTestWebhookNotification())

    async def get_webhook_secret_key(self, hook_id: str) -> str:
        """
        Each notification contains a digital signature of the message, encrypted with a key.
        To obtain a signature verification key, use this request.

        :param hook_id: UUID of webhook
        :return: Base64 encoded key
        """
        return await self._request_service.execute_api_method(GetWebhookSecret(hook_id=hook_id))

    async def delete_current_webhook(self) -> Dict[Any, Any]:
        """Method to delete по webhook"""
        webhook = await self.get_current_webhook()
        return await self._request_service.execute_api_method(DeleteWebhook(hook_id=webhook.id))

    async def generate_new_webhook_secret(self, hook_id: str) -> str:
        """
        Use this request to change the encryption key for notifications.

        :param hook_id: UUID of webhook
        :return: Base64 encoded key
        """
        return await self._request_service.execute_api_method(
            GenerateWebhookSecret(hook_id=hook_id)
        )

    async def bind_webhook(
        self,
        url: Union[str, WebhookURL],
        *,
        transactions_type: int = 2,
        send_test_notification: bool = False,
        delete_old: bool = False,
    ) -> Tuple[WebhookInfo, str]:
        """
        [NON-API] EXCLUSIVE method to register new webhook or get old

        :param url: service endpoint
        :param transactions_type: 0 => incoming, 1 => outgoing, 2 => all
        :param send_test_notification:  test_qiwi will transfer_money
         you test webhook update
        :param delete_old: boolean, if True - delete old webhook

        :return: Tuple of Hook and Base64-encoded key
        """
        if isinstance(url, WebhookURL):
            url = url.render()

        if delete_old:
            with suppress(QiwiAPIError):
                await self.delete_current_webhook()

        webhook = await self.register_webhook(url, transactions_type)
        key = await self.get_webhook_secret_key(webhook.id)

        if send_test_notification is True:
            await self.send_test_webhook_notification()

        return webhook, key

    @property
    def phone_number_without_plus_sign(self) -> str:
        if self._phone_number is None:
            raise RuntimeError('Phone number is empty')
        return self._phone_number[1:]



from pydantic import Field

from glQiwiApi.types.base import HashableBase


class Restriction(HashableBase):
    code: str = Field(..., alias='restrictionCode')
    description: str = Field(..., alias='restrictionDescription')


__all__ = ('Restriction',)



from pydantic import Field

from glQiwiApi.types.amount import AmountWithCurrency
from glQiwiApi.types.base import Base


class Commission(Base):
    provider_id: int = Field(alias='providerId')
    withdraw_sum: AmountWithCurrency = Field(alias='withdrawSum')
    enrollment_sum: AmountWithCurrency = Field(alias='enrollmentSum')
    qiwi_commission: AmountWithCurrency = Field(alias='qwCommission')
    withdraw_to_enrollment_rate: int = Field(alias='withdrawToEnrollmentRate')
